#ifndef SKIPROPE_H_INCLUDED
#define SKIPROPE_H_INCLUDED


//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.




#include <stdlib.h>
#include <new>
#include <vector>
#include <portable_io.h>

namespace cxxtls
{


///@file
///
///  This file defines the skiprope container class.
/// 
///  A skiprope\<T> is very large vector of T's which can be iterated over
///  and which has ln(N) insertion, removal, and positioning characteristics.
///  Once an iterator is found for a given position within the vector, inc'ing
///  and dec'ing it are O(1), but Adding a specific amount to it is O(N).
/// 
///  The memory utilization of an empty skiprope<T> is quite high, so don't bother
///  using this container if you need to save every byte of memory -- it is meant
///  to give fast inserts and deletes in large vector situations.
/// 
///  Note that every insert or erase to the skiprope invalidates all iterators
///  into the rope.  This is a vector, not a list.

#ifndef SKIPROPE_MAX_BLOCKSIZE
#  define SKIPROPE_MAX_BLOCKSIZE  (32000) /* bytes, change only for testing */
#endif

template<class T> class skiprope_debugger;

template<class T> class skiprope_const_iterator;
template<class T> class skiprope_iterator;
template<class T> class skiprope;

//
//  A skiprope is an array of pointers to node lists.  The skiprope has
//  approximately vector semantics but instead of random accesses being O(1)
//  they are O(N).  See class skiprope below for more complexity numbers
//
//  All of the nodes appear in array index[0] list.  About 1/6 appear in
//  index[1] list, about 1/36 appear in index[2] list, etc.  This allows for
//  binary searches to find a given address within the simulated vector.
//



template<class T>
class skiprope_node
/// 
///   A skiprope_node is a chunk of a skiprope<T>.  Each node is a vector<T>
///   and the number of T's in the vector varies from node to node.  The
///   start position of any given node is equal to the sum of the nodes before
///   it.
/// 
///   skiprope nodes vary in the size of their data structure.  Each node has
///   between 1 and 32 (or 48, or whatever -- the algorithm behaves well
///   regardless of the maximum height -- but it mustn't bee too small
///   or you'll get lots of searching.  Too tall and you get lots of wasted
///   space (well, if T is large and the vector chunk size is large you won't) ).
/// 
///   Skiprope nodes appear in multiple linked lists (and they have linkage info
///   for each in which they appear).  The linkage is bidirectional and each node
///   level's linkage information contains a width_ member which is the distance
///   from this node, to the next_ node at this level).  Here is a picture showing
///   how nodes are linked together in the skiprope and what width_ members contain:
///   @code
/// 
///     lvl  Nodes/levels
///     ---+--------------------------
///       3| X
///       2| X     Z
///       1| X     Z           H   J
///       0| X A B Z C D E F G H I J
/// 
///     Here is a small portion of the linkage data:
/// 
///       X[0].next_  = A
///       X[0].width_ = width(X)
/// 
///       A[0].prev_  = X
///       A[0].width_ = width(A)
/// 
///       B[0].prev_  = A
///       B[0].next_  = Z
///       B[0].width_ = width(B)
///       ...
/// 
///       X[1].next_ = Z
///       X[1].width_= width(X) + width(A) + width(B)
///       <A and B have only level 0 info>
///       ...
///       Z[1].next_   =  H
///       Z[1].prev_   =  X
///       Z[1].width_  =  width(Z) + width(D) + width(E) +
///                       width(F) + width(G)
///       ...
///       X[2].next_   = Z
///       X[2].width_  = width(X) + width(A) + width(B)
///       Z[2].next_   =  H
///       Z[2].prev_   =  X
///       Z[2].width_  =  width(Z) + width(D) + width(E) +
///                       width(F) + width(G)
///       ...
///       X[3].width_  = width(X)
/// 
///   @endcode
{
  int levels_;       ///< number of linkage levels in this node

  std::vector<T> data_;   ///< data stored in this node

  typedef typename std::vector<T>::iterator iterator;
  typedef typename std::vector<T>::const_iterator const_iterator;

  // array link_ is defined at the bottom of this structure see note LAST
  // about it below

  friend class skiprope<T>;
  friend class skiprope_const_iterator<T>;
  friend class skiprope_iterator<T>;

  friend class skiprope_debugger<T>; // testing tool

  struct link_info
  {
    skiprope_node<T> *next_;
    skiprope_node<T> *prev_;
    ptrdiff_t         width_; ///< width of lower level successors to this node
			      ///  between this node and next_. (width is sum
			      ///  of this->data_.size() plus the some of the
			      ///  size() of intervening nodes at each level
			      ///  below the this level.  See insert_node
			      ///  below so as to understand the logic.
  };

public:

  typedef skiprope_node<T> self;

  iterator       begin()        { return data_.begin(); }
  const_iterator begin() const  { return data_.begin(); }

  iterator       end()          { return data_.end(); }
  const_iterator end() const    { return data_.end(); }

  void insert(T const &d, ptrdiff_t o);
  void insert(T const *f, T const *l, ptrdiff_t o);
  #ifdef _AIX
  void insert(const_iterator f, const_iterator l, ptrdiff_t o);
  #endif
  void push_back(T const *f, T const *l);
  void erase(ptrdiff_t o);
  void erase(ptrdiff_t o, ptrdiff_t cnt);
  void erase(typename std::vector<T>::iterator b, typename std::vector<T>::iterator e);

  void adjust_widths(ptrdiff_t); // adjust link_[*].width_

  ptrdiff_t size() const   { return data_.size(); }
  int       levels() const { return levels_; }

  static self * new_node(int max_possible_level)
  ///
  ///  allocate a new node which will have a level which is some random
  ///  number between 1 and max_possible_level, inclusive.  Space will be
  ///  allocated in the returned pointer for the link_ array to be treated
  ///  as link_[levels_].
  ///
  {
    // pick the level

    int l = 1;

    while(l < max_possible_level)
    {
      int v = rand();
      int m = RAND_MAX/3;

      if(v > m)
	 break;
      ++l;
    }

    // allocate space

    self *rv = (self*)(new char[ sizeof(self) + (l-1)*sizeof(link_info) ]);

    // construct the data members

    rv->levels_ = l;

    new (&rv->data_) std::vector<T>;

    for(int i=0; i < l; ++i)
    {
      rv->link_[i].next_  = 0;
      rv->link_[i].prev_  = 0;
      rv->link_[i].width_ = 0;
    }

   return rv;
  }

  static void discard(self *victim)
  {
    typedef std::vector<T> MyVector;  // overcome portability issues 

    MyVector *p = &victim->data_;

    p->MyVector::~MyVector();

    delete [] (char*)(victim);

  }

private:
  void *operator new(size_t);   // not implemented, don't use this
  void operator delete(void*p); // not implemented, don't use this

//NOTE(LAST);  The following data structure must be the last piece of data in this structure
//

  link_info link_[1]; // really contains levels_ elements

// STOP!  See note, "LAST", above
//        Put _no_ data below this line
}; // put no data above this line

template<class T>
void skiprope_node<T>::adjust_widths(ptrdiff_t amount)
//
///   adjust the width_ members of this node's link info to account
///   for inserts or deletes in data_
//
{
  link_info *first = &link_[0];
  link_info *last  = first + levels_;

  while(first != last)
  {
    first->width_ += amount;
    ++first;
  }

}


template<class T>
void skiprope_node<T>::insert(T const &d, ptrdiff_t o)
{
  data_.insert(data_.begin()+o,d);
  adjust_widths(o);
}

template<class T>
void skiprope_node<T>::insert(T const *f, T const *l, ptrdiff_t o)
{
  data_.insert(data_.begin()+o, f, l);
  adjust_widths(l - f);
}

#ifdef _AIX
template<class T>
void skiprope_node<T>::insert(typename skiprope_node<T>::const_iterator f, 
                              typename skiprope_node<T>::const_iterator l, 
                              ptrdiff_t o)
{
  data_.insert(data_.begin()+o, f, l);
  adjust_widths(l - f);
}
#endif


template<class T>
void skiprope_node<T>::push_back(T const *f, T const *l)
{
  data_.insert(data_.end(),f,l);
  adjust_widths(l - f);
}

template<class T>
void skiprope_node<T>::erase(ptrdiff_t o)
{
  data_.erase(data_.begin()+o, data_.begin()+(o+1));
  adjust_widths(-1);
}
template<class T>
void skiprope_node<T>::erase(ptrdiff_t o, ptrdiff_t cnt)
{
  data_.erase(data_.begin()+o, data_.begin()+(o+cnt));
  adjust_widths(-cnt);
}
template<class T>
void skiprope_node<T>::erase(typename std::vector<T>::iterator b,
			     typename std::vector<T>::iterator e)
{
  data_.erase(b,e);
  adjust_widths(-(e - b));
}

template<class T>
class skiprope_iterator
//
///   A skiprope iterator simulates a pointer to type T.  The iterator knows
///   about the structure of a skiprope.  It knows that a skiprope is a linked
///   list of nodes of varying widths.
//
{
  typedef skiprope<T> rope;

  skiprope_node<T> *node_;
  ptrdiff_t         offset_;
  rope             *rope_;  // needed to get end() when out of data

  friend class skiprope_const_iterator<T>;
  friend class skiprope<T>;
  friend class skiprope_debugger<T>;       // testing tool

public:

  typedef skiprope_iterator<T> self;

  skiprope_iterator(): node_(0), offset_(0), rope_(0) {}

  skiprope_iterator(self const &r): node_(r.node_), offset_(r.offset_), rope_(r.rope_) {}

  skiprope_iterator(skiprope_node<T> *n, ptrdiff_t o, skiprope<T> *r, int normal=1);

  T &operator*()             { return *(node_->begin() + offset_); }
  T const &operator*() const { return *(node_->begin() + offset_); }

  // The begin and end methods provide a way of finding out the range
  // of data quickly addressable by this node.  In other words, an iterator
  // has fast access to the data in the node specified during the iterator's
  // creation.  If you wish to make use of this faster access, you can use the
  // begin and end methods to get iterators to the beginning and ending thereof.

  self begin() { return self(node_,0); }
  self end()   { return self(node_, node_->data_.size()); }

  self const begin() const { return self(node_,0); }
  self const end()   const { return self(node_, node_->data_.size()); }

  int operator==(self const &r) const
    {
      return node_ == r.node_ && offset_ == r.offset_;  // ignore rope_
    }

  int operator!=(self const &r) const
    {
      return node_ != r.node_ || offset_ != r.offset_;
    }

  self &operator=(self const &r)
    {
      node_   = r.node_;
      offset_ = r.offset_;
      rope_   = r.rope_;
      return *this;
    }

  self &operator++();
  self  operator++(int);
  self &operator--();
  self  operator--(int);

};

template<class T>
skiprope_iterator<T>::
skiprope_iterator(skiprope_node<T> *n, ptrdiff_t o, skiprope<T> *r, int normal)
: node_(n), offset_(o), rope_(r)
{
  if(node_ == 0)
  {
    offset_ = 0;
    return;
  }

  if(!normal)
  {
    return;  // special cases where searching to the next node is not
	     // desireable -- used only by skiprope<> to jam in iterators
	     // that aren't normally valid
  }

  int s = n->size();

  while(node_ && offset_ >= s)
  {
    offset_ -= s;
    node_ = node_->link_[0].next_;
    if(node_)
      s = node_->size();
  }
  if(node_ == 0)
  {
    node_ = rope_->last_;

    if(node_)
    {
      offset_ = node_->size();
    }
    else
    {
      offset_ = 0;
    }
  }
}


template<class T>
class skiprope_const_iterator
//
///   A skiprope iterator simulates a pointer to type const T.  The iterator knows
///   about the structure of a skiprope.  It knows that a skiprope is a linked
///   list of nodes of varying widths.
//
{
  typedef skiprope<T>     rope;
  skiprope_node<T> const *node_;
  ptrdiff_t               offset_;
  rope const             *rope_;

  friend class skiprope<T>;
  friend class skiprope_debugger<T>; // testing tool

public:

  typedef skiprope_const_iterator<T> self;

  skiprope_const_iterator(): node_(0), offset_(0), rope_(0) {}

  skiprope_const_iterator(self const &r)
  : node_(r.node_),
    offset_(r.offset_),
      rope_(r.rope_)
  {
  }

  skiprope_const_iterator(skiprope_node<T> const *n, ptrdiff_t o, rope const* r);

  skiprope_const_iterator(skiprope_iterator<T> i)
  : node_(i.node_),
    offset_(i.offset_),
    rope_(i.rope_)
  {
  }

  self &operator=(self const &r)
    {
      node_   = r.node_;
      offset_ = r.offset_;
      rope_   = r.rope_;
      return *this;
    }

  self &operator=(skiprope_iterator<T> const &r)
    {
      node_   = r.node_;
      offset_ = r.offset_;
      rope_   = r.rope_;
      return *this;
    }


  T const &operator*() const { return *(node_->begin() + offset_); }

  // The begin and end methods provide a way of finding out the range
  // of data quickly addressable by this node.  In other words, an iterator
  // has fast access to the data in the node specified during the iterator's
  // creation.  If you wish to make use of this faster access, you can use the
  // begin and end methods to get iterators to the beginning and ending thereof.


  self begin() { return self(node_,0); }
  self end()   { return self(node_, node_->data_.size()); }

  self const begin() const { return self(node_,0); }
  self const end()   const { return self(node_, node_->data_.size()); }

  int operator==(self const &r) const
    {
      return node_ == r.node_ && offset_ == r.offset_;
    }

  int operator!=(self const &r) const
    {
      return node_ != r.node_ || offset_ != r.offset_;
    }

  int operator==(skiprope_iterator<T> const &r) const
    {
      return node_ == r.node_ && offset_ == r.offset_;
    }

  int operator!=(skiprope_iterator<T> const &r) const
    {
      return node_ != r.node_ || offset_ != r.offset_;
    }

  self &operator++();    // pre-increment
  self operator++(int);  // post-increment
  self &operator--();    // pre-decrement
  self operator--(int);  // post-decrement

};

template<class T>
skiprope_const_iterator<T>::
skiprope_const_iterator(skiprope_node<T> const *n, ptrdiff_t o, rope const* r)
: node_(n),
  offset_(0),
  rope_(r)
{
  if(node_ == 0)
  {
    offset_ = 0;
    return;
  }

  ptrdiff_t s = node_->size();

  while(node_ && offset_ >= node_->size())
  {
    offset_ -= s;
    node_ = node_->link_[0].next_;
  }
  if(node_ == 0)
  {

    node_ = rope_->last_;

    if(node_)
    {
      offset_ = node_->size();
    }
    else
    {
      offset_ = 0;
    }

  }
}


template<class T>
struct skiprope_level_desc
{
  typedef skiprope_node<T> node;

  node       *head_;    // first element at this level
  ptrdiff_t   offset_;  // to element at head_ from beginning of skiprope

  skiprope_level_desc(): head_(0) {}

};

template<class T>
class skiprope
/// 
///   An vector-like collection<T> which is implemented in such a way that
///   no large blocks of memory are created.  This allows inserts and deletes to
///   be  done in a time which is proportional to the size of the block being
///   acted upon -- rather than upon the whole size of the container.  To pay
///   for _this_ performance gain, random accesses within the vector require
///   O(ln(N)) time rather than O(1) as you usually see in a vector.  Iteration
///   is still O(1) but is much slower than iteration in a vector<T>.
/// 
///   The following table describes the time complexity of some normal operations:
/// 
///   @code
///     skiprope()     O(1)        // create an empty skiprope
///    ~skiprope()     O(N)        // destroy a skiprope
///     operator[i]    O(ln(N))    // get a reference to a specifc location
///     find(i)        O(ln(N))    // get an iterator to a specific location
///     iterator++     O(1)        // increment an iterator
///     iterator--     O(1)        // decrement an iterator
///     iterator+int   --------    // can't add numbers to iterators, use find()
///     insert(it,T)   O(ln(N))    // insert given an iterator
///     erase(it)      O(ln(N))    // erase given an iterator
///     push_back(T)   O(1)        // append
///   @endcode
///   This container is implemented as a sorted list of variable width blocks.
///   The sum of the width's of the prior blocks is the start offset of this block.
///   When inserting data into a block, no block can exceed max_block_size
///   unless sizeof(T) is greater than max_block_size.  The fact that the blocks
///   are sorted means that you can
/// 
///   Each node of data (again, varying in width between nodes) is assigned
///   1 to max_index_levels randomly.  The randomization gives the O(Ln(N))
///   behavior with small wasted space in the indices.  Nodes are linked into
///   at most max_index_levels of linked lists based on the level assigned to
///   the particular node.
/// 
///   To get O(LN(N)) lookups for a given offset into the vector, the highest
///   level index list is searched first. Then the list of nodes that might hold
///   the offset in question is found.
/// 
///   Next, the next lower level of lists is searched to narrow the range.  The
///   narrowed range is then further narrowed by going down one more level
///   of lists.  This is repeated until the 0'th level is searched.
/// 
///   This simulates a binary tree but is always balanced -- ergo no rebalancing
///   penalty is ever paid.
/// 

{
  friend class skiprope_debugger<T>; // testing tool
  friend class skiprope_iterator<T>; // needed for implementations below
  friend class skiprope_const_iterator<T>; // needed for implementations below


  enum max_block_size_ { max_block_size=SKIPROPE_MAX_BLOCKSIZE };

  enum index_levels_ { max_index_levels=32 };

  enum chunk_size_   { chunk_size =                   ///< T's per max_block_size
			 (sizeof(T) > max_block_size
			   ?
			    1
			   :
			   max_block_size / sizeof(T)
			 )
		     };

  typedef skiprope_node<T> node;
  typedef typename skiprope_node<T>::link_info  link_info;

  typedef skiprope_level_desc<T> level_desc;

  level_desc level_[max_index_levels];


  node     *last_;           ///< ptr to last node in the rope
			     ///  gives O(1) rear access and makes end()
			     ///  work correctly

  int       active_levels_;  ///< hightest node level

  ptrdiff_t size_;           ///< items in the array

  int       nodes_;

public:

  int       visited_levels_; ///< debug info for test only.

  typedef skiprope<T> self;
  typedef int	      size_t;

  typedef skiprope_iterator<T> iterator;

  typedef skiprope_const_iterator<T> const_iterator;

  skiprope(): last_(0), active_levels_(0), size_(0), nodes_(0) {}
  skiprope(self const &r);

  ~skiprope();

  ptrdiff_t size()  const { return size_; }

  iterator  begin();
  iterator  end();
  iterator  find(ptrdiff_t offset);   ///< iterator to offset
  ptrdiff_t find(iterator);           ///< offset from iterator
  T&        operator[] (ptrdiff_t i)       { return *find(i); }


  const_iterator begin()  const;
  const_iterator end()    const;
  const_iterator find(ptrdiff_t offset) const;
  ptrdiff_t      find(const_iterator) const;
  T const&       operator[] (ptrdiff_t i) const { return *find(i); }

  iterator insert(iterator, T const* first, T const * last);
  iterator insert(iterator i, T const &r);
  void erase(iterator location);
  void erase(iterator first, iterator last);

  void push_back(T const *first, T const *last);
  void push_back(T const &r);
  iterator push_front(T const *first, T const *last);
  iterator push_front(T const &r);
  T&       front()        { return *begin(); }
  T const& front() const  { return *begin(); }
  T&       back()         { iterator t = end(); --t; return *t; }
  T const& back()  const  { const_iterator t = end(); --t; return *t; }


private:

  void insert_node(node *before, node *after); // adjusts size_

  void split_node(node *n, ptrdiff_t offset);  // cuts node into 2 pieces
					       // and adjusts


public:
  struct node_info
      /// A structure describing a node in the rope.  It is used in
      /// navigating the node data structure more easily than keeping track
      /// of the separate components of the description.
  {
    node      *node_;
    ptrdiff_t  offset_; ///< from begininng to node

    node_info(node*n, ptrdiff_t o): node_(n), offset_(o) {}

    node_info(){}

  };

  friend struct node_info;

private:

  node_info find_node_before(ptrdiff_t location, int level);



   int predecessors(node *n, node_info *pred);  ///< get list of pred nodes
						///  returns count of levels
						///  returned.

  int prednodes(node *, node_info *pred);  ///< get only node_ info into pred
					   ///  and return count of levels found
  void predoffsets(node_info*pred, int levels);  ///< compute offset_ given nodes_

};

template<class T>
skiprope<T>::~skiprope()
{
  node *scan = level_[0].head_;

  while(scan)
  {
    node *next = scan->link_[0].next_;

    node::discard(scan);

    scan = next;

  }

}

template<class T>
skiprope<T>::skiprope(typename skiprope<T>::self const &r)
: last_(0),
  active_levels_(0),
  size_(0)
///
///  copy construct from another rope
///
{
  node *scan = r.level_[0].head_;

  while(scan)
  {
    push_back(scan->begin(), scan->end());

    scan = scan->link_[0].next_;
  }

}

template<class T>
typename skiprope<T>::iterator
skiprope<T>::begin()
{
  return iterator(level_[0].head_, 0, this);
}

template<class T>
typename skiprope<T>::const_iterator
skiprope<T>::begin() const
{
  return const_iterator(level_[0].head_, 0);
}

template<class T>
typename skiprope<T>::iterator
skiprope<T>::end()
{
  return iterator(last_, last_?last_->size():0, this);
}

template<class T>
typename skiprope<T>::const_iterator
skiprope<T>::end() const
{
  return const_iterator(last_, last_?last_->size():0);
}

template<class T>
void skiprope<T>::push_back(T const &r)
{
  push_back(&r, &(r)+1);
}

template<class T>
typename skiprope<T>::iterator
skiprope<T>::insert(typename skiprope<T>::iterator where,  T const &r)
{
  return insert(where, &r, &(r)+1);
}


template<class T>
typename skiprope<T>::iterator
skiprope<T>::find(ptrdiff_t offset)
/// 
///  Get an iterator to a specified offset within the skiprope vector
/// 
///  starting at the highest active level
///  scan across the links on each level until the sum of the
///  widths of the links already scanned plus the width of the next
///  link will exceed the search address.  When the next link exceeds
///  or doesn't exist, drop down a level at the same node and keep scanning until
///  you are at level 0 or found it or out of data.  The scanner moves
///  from the top left of a graph to the bottom right where top left is
///  the highest array index of the level_ array and offset 0 from the
///  beginning of the simulated vector.
/// 
///  Each node has a width_ for each level.  This width is the sum of distance
///  of all lower level nodes between this node and the next node _this_ level.
///  At any given level, the start address within the simulated vector of this
///  node is the sum of width_ fields at this level prior (ie between head_
///  and this node).

{ // find

  int level;
  node *n;
  ptrdiff_t p=0;

  for(level = active_levels_-1; level >= 0 && offset < (p=level_[level].offset_); --level);

  if(level >= 0)
  {
    // found a row containing the desired offset

    n = level_[level].head_;

    while(level >= 0)
    {
      while(n->link_[level].next_ &&
	    offset >= (p + n->link_[level].width_)
	   )
      {
	p += n->link_[level].width_;
	n  = n->link_[level].next_;
      }

      if(offset < p + n->size())
      {
	 return iterator(n, offset - p, this);
      }

      --level;
      ++visited_levels_;  // statistics for tests, etc.
    }


  }

  return end();

} // find

template<class T>
typename skiprope<T>::node_info
skiprope<T>::find_node_before(ptrdiff_t offset, int search_level)
/// 
///   get the node just before 'offset' at a given level
/// 

{ // find_node_before

  int level;
  node *n;
  ptrdiff_t p=0;

  for(level = active_levels_-1; level >= search_level && offset < (p=level_[level].offset_); --level);

  if(level >= 0)
  {
    // found a row containing the desired offset

    n = level_[level].head_;

    while(level >= 0)
    {
      while(n->link_[level].next_ &&
	    offset >= (p + n->link_[level].width_)
	   )
      {
	p += n->link_[level].width_;
	n  = n->link_[level].next_;
      }

      if(offset < p + n->size())
      {
	 return node_info(n,p);
      }

      --level;
    }

  }

  return node_info(0,0);

} // find_node_before


template<class T>
typename skiprope<T>::const_iterator
skiprope<T>::find(ptrdiff_t offset) const
{
  self *unconst = this;  // cast away const

  return unconst->find(offset);
}

template<class T>
void
skiprope<T>::push_back(T const *first, T const *last)
//
///   append a bunch of T's to the end of the skiprope
//

{ // push_back

  ptrdiff_t insert_count = last - first;

  if(last_)
  { // use empty space in extant last node


    ptrdiff_t avail = chunk_size - last_->size();

    if(avail > insert_count)
    {
      last_->push_back(first, last);
      size_ += insert_count;
      return;
    }
    else
    {
      last_->push_back(first, first+avail);
      first += avail;
      insert_count -= avail;
      size_ += avail;
    }
  } // use empty space in extant last node

  // if we get here, insert_count and first will have been
  // adjusted so that we must attend to only that which remains
  // to be done

  while(insert_count)
  { // append new nodes to end

    node *n = node::new_node(max_index_levels); // nodes_ is inc'd later
    ptrdiff_t max_insert = insert_count;

    if(max_insert > chunk_size)
    {
      max_insert = chunk_size;
    }

    n->push_back(first, first+max_insert);

    insert_count -= max_insert;

    first += max_insert;

    if(last_)
    {
      insert_node(last_, n);
    }
    else
    {
      // inserting the first node

      last_ = n;

      size_ += max_insert;
      ++nodes_;

      int i;

      active_levels_ = n->levels();

      for(i=0; i < n->levels(); ++i)
      {
	level_[i].head_   = n;
	level_[i].offset_ = 0;
      }


    }

  } // append new nodes to end

} // push_back

template<class T>
int skiprope<T>::
prednodes(typename skiprope<T>::node *n,
	  typename skiprope<T>::node_info *pred
	 )
{
  int level=0;

  while(level < active_levels_)
  {
    //
    //  copy the current node's info out
    //
    while( level < n->levels() )
    {
      pred[level].node_ = n;
      ++level;
    }

    // if all levels accounted for then quit

    if(level >= active_levels_)
      break;

    // otherwise scan backwards along the level of linkage
    // in node to find prev guy (this still maintains O(ln(N))
    // behavior because we are stepping to ever higher levels
    // as soon as one is found -- thus jumping over lower level
    // nodes.

    node *old_n = n;

    while(n->link_[level-1].prev_ &&
	  n->levels() <= level
	 )
    {
      if(n->levels() < level )
      {
	printf("prev(%p) points to lower level node\n", n);
	exit(1);
      }
      n = n->link_[level-1].prev_;
    }

    if( ( old_n == n || (n == 0 && level != active_levels_) ) )
    {
      break;
    }

 }

 return level;

}

template<class T>
void skiprope<T>::
predoffsets(typename skiprope<T>::node_info *pred, int level)
  //
  /// given node_info with the nodes_ filled in, compute the offsets_ from
  /// the beginning of the pseudo vector for each node.  The assumption is
  /// that the pred array was created by function prednodes().  Any other
  /// data will just malfunction horribly.
  //
{
  --level;

  ptrdiff_t offset = level_[level].offset_;

  node *n = level_[level].head_;

  while(level >= 0)
  {
    // scan across then down the levels

    while(n->link_[level].next_ && n != pred[level].node_ )
    {
      // scan across this level until we find the pred[] entry

      offset += n->link_[level].width_;
      n = n->link_[level].next_;
    }

    pred[level].offset_ = offset;

    --level;

    while(level >= 0 && pred[level].node_ == n)
    {
      pred[level].offset_ = offset;
      --level;
    }


  }

}


template<class T>
int skiprope<T>::predecessors(typename skiprope<T>::node *n,
			       typename skiprope<T>::node_info *pred
			      )
//
///   Get a list of pointers to predecessor node information which
///   must be adjusted whenever you are inserting a new node after 'n'.
///   'node' will be included in the output pointer array, pred.  That is,
///   the purpose of this routine is to get get the predecessors of a node
///   AFTER 'node', not OF node.
/// 
///   The node information consists of pointers and offsets from the beginning
///   of the pseudo vector _TO_ the node.
//
{
  // first, get the node pointers

  int level=prednodes(n, pred);


  // then get the offset_'s by scanning along the the horizontal links
  // and adding up the distance between jumps.  When there is no next
  // node at the level, drop down and scan at that level

  predoffsets(pred, level);

  return level;

}

template<class T>
void skiprope<T>::
insert_node(typename skiprope<T>::node *before,
	    typename skiprope<T>::node *after
	   )
//
///   update the skiprope info to include node 'after' immediately after 'before'.
/// 
///   Here is a typical scenario:
///   @code
/// 
///     lvl Nodes             In this case, after, represented by 'A' is
///     ---+------------      less in height than before, represented by 'B'.
///      4|                   But the highest predecessor level is even greater
///      3|           P       than the height of 'B' (its 2 -- associated with 'N').
///      2|   N       P       And both are less than the highest active level (3 --
///      1|   N   B   P       associated with 'P').
///      0| M N O B A P
///                 ^
///@endcode
{


  ptrdiff_t after_size = after->size();

  size_ += after_size;

  if(last_ == before)
  {
    last_ = after;
  }

  node_info pred[max_index_levels];

  int pred_level = predecessors(before, pred);

  int after_levels = after->levels();

  ptrdiff_t after_location = pred[0].offset_ + pred[0].node_->size();

  int i;

  ++nodes_;

  for(i=0; i < pred_level; ++i)
  {
    // link after to its predecessors in those areas
    // common to the before and after node

    if( i >= after_levels )
    {
      pred[i].node_->link_[i].width_ += after_size;
      continue;
    }

    ptrdiff_t pred_width;
    ptrdiff_t after_width;

    pred_width =  after_location  - pred[i].offset_;

    if(pred[i].node_->link_[i].next_)
    {
      after_width=  pred[i].offset_ + pred[i].node_->link_[i].width_ +
		    after_size - after_location;
    }
    else
    {
      after_width = after_size;
    }

    after->link_[i].next_ = pred[i].node_->link_[i].next_;
    after->link_[i].prev_ = pred[i].node_;
    pred[i].node_->link_[i].next_  = after;

    after->link_[i].width_         = after_width;
    pred[i].node_->link_[i].width_ = pred_width;

    if(after->link_[i].next_)
    {
      after->link_[i].next_->link_[i].prev_ = after;
    }

  }

  while( i < after_levels && i < active_levels_ )
  {
    // if the after node is higher than all its predecessors
    // link the after node into the chains of level_ array
    // Here is the kind of thing I'm talking about:
    //
    //    lvl Nodes
    //    ---+------------
    //     4|
    //     3|             P
    //     2|       X     P
    //     1|   N   X B   P
    //     0| M N O X B A P
    //              ^
    //  Here, X is inserted after O and it's levels are higher than those
    //  of any of its predecessors.  Thus, it must be linked into the
    //  level 2 horizontal chain prior to P.
    //

    if(after_location < level_[i].offset_)
    {
      // this guy becomes the first link in the level_ horizontal chain

      after->link_[i].width_ = level_[i].offset_ - after_location + after_size;
      after->link_[i].next_  = level_[i].head_;

      if(level_[i].head_)
        level_[i].head_->link_[i].prev_ = after;

      level_[i].offset_      = after_location;
      level_[i].head_        = after;
    }
    else
    {
      //
      //  In this case, after, has a higher number of levels than before and all of its
      //  predecessors.  We must link the new node into the level_[].  Note that this
      //  new node will be the first node at this level -- otherwise, predecessors nodes
      //  would have taken it into acount
      //

      after->link_[i].width_ = after_size + level_[i].offset_ - after_location;
      after->link_[i].next_  = level_[i].head_;
      level_[i].offset_      = after_location;
      level_[i].head_        = after;

      if(after->link_[i].next_)
      {
	after->link_[i].next_->link_[i].prev_ = after;
      }

    }

    ++i;

  }

  while(i < active_levels_ )
  {
    // push everyone else out
    level_[i].offset_ += after_size;
    ++i;
  }

  while(active_levels_ < after_levels)
  {
    // add new levels as needed

    level_[active_levels_].head_   = after;
    level_[active_levels_].offset_ = after_location;
    ++active_levels_;
  }

}

template<class T>
skiprope_iterator<T> &
skiprope_iterator<T>::
operator++()  ///  pre-increment
{
  ++offset_;

  if(offset_ >= node_->size() )
  {
    offset_ = 0;

    node_ = node_->link_[0].next_;

    if(node_ == 0)
      *this = rope_->end();
  }
  return *this;
}

template<class T>
skiprope_iterator<T>
skiprope_iterator<T>::
operator++(int) ///  post increment
{
  self rv(*this);

  ++offset_;

  if(offset_ >= node_->size() )
  {
    offset_ = 0;

    node_ = node_->link_[0].next_;

    if(node_ == 0)
      *this = rope_->end();

  }

  return rv;
}

template<class T>
skiprope_iterator<T> &
skiprope_iterator<T>::
operator--()  ///  pre-decrement
{
  if(*this == rope_->end())
  {
    node_ = rope_->last_;

    if(node_)
      offset_ = node_->size() -1;
    else
      offset_ = 0;

    // this assumes that no node will be 0 length and not be
    // erased.

  }
  else
  {
    if(offset_ == 0)
    {
      node_ = node_->link_[0].prev_;

      if(node_ != 0)
        offset_ = node_->size() - 1;

       
    }
    else
    {
      --offset_;
    }
  }
  return *this;
}

template<class T>
skiprope_iterator<T>
skiprope_iterator<T>::
operator--(int) ///  post decrement
{
  self rv(*this);

  if(*this == rope_->end())
  {
    node_ = rope_->last_;

    if(node_)
      offset_ = node_->size() -1;
    else
      offset_ = 0;

    // this assumes that no node will be 0 length and not be
    // erased.

  }
  else
  {
    if(offset_ == 0)
    {
      node_ = node_->link_[0].prev_;
      if(node_)
        offset_ = node_->size() - 1;
    }
    else
    {
      --offset_;
    }
  }


  return rv;
}

//=============================================================================
template<class T>
skiprope_const_iterator<T> &
skiprope_const_iterator<T>::
operator++()  ///  pre-increment
{
  ++offset_;

  if(offset_ >= node_->size() )
  {
    offset_ = 0;

    node_ = node_->link_[0].next_;

    if(node_ == 0)
      *this = rope_->end();
  }
  return *this;
}

template<class T>
skiprope_const_iterator<T>
skiprope_const_iterator<T>::
operator++(int) ///  post increment
{
  self rv(*this);

  ++offset_;

  if(offset_ >= node_->size() )
  {
    offset_ = 0;

    node_ = node_->link_[0].next_;

    if(node_ == 0)
      *this = rope_->end();

  }

  return rv;
}

template<class T>
skiprope_const_iterator<T> &
skiprope_const_iterator<T>::
operator--()  ///  pre-decrement
{
  if(*this == rope_->end())
  {
    node_ = rope_->last_;

    if(node_)
      offset_ = node_->size() -1;
    else
      offset_ = 0;

    // this assumes that no node will be 0 length and not be
    // erased.

  }
  else
  {
    if(offset_ == 0)
    {
      node_ = node_->link_[0].prev_;
      if(node_)
        offset_ = node_->size() - 1;
    }
    else
    {
      --offset_;
    }
  }
  return *this;
}

template<class T>
skiprope_const_iterator<T>
skiprope_const_iterator<T>::
operator--(int) ///  post decrement
{
  self rv(*this);

  if(*this == rope_->end())
  {
    node_ = rope_->last_;

    if(node_)
      offset_ = node_->size() -1;
    else
      offset_ = 0;

    // this assumes that no node will be 0 length and not be
    // erased.

  }
  else
  {
    if(offset_ == 0)
    {
      node_ = node_->link_[0].prev_;

      if(node_)
        offset_ = node_->size() - 1;
    }
    else
    {
      --offset_;
    }
  }


  return rv;
}

template<class T>
typename skiprope<T>::iterator
skiprope<T>::push_front(T const &r)
{
  return push_front(&r, (&r)+1);
}


template<class T>
typename skiprope<T>::iterator
skiprope<T>::push_front(T const *first, T const *last)
  //
  ///  When inserting at the front, it is easier than most cases
  ///  because every node is already in level 0.  So, for example if
  ///  you have the following layout:
  ///
  ///  @code
  ///    lvl  Nodes
  ///    ---+---------------
  ///     3|         E
  ///     2|       D E
  ///     1| A   C D E
  ///     0| A B C D E
  ///  @endcode
  ///  Inserting at the front, if the data fits in A,  means doing the following:
  ///
  ///  @code
  ///    Adjusting the width of A (from 1 to 2, for example)
  ///
  ///    adjusting the level_[].offset_ info for all nodes
  ///      who have levels higher than A's:  D and E.
  ///
  ///  @endcode
  ///  On the other hand, if the data won't fit in the first node, we must
  ///  insert a new one:
  ///
  ///  @code
  ///    lvl  Nodes
  ///    ---+---------------
  ///     3|           E
  ///     2| N       D E
  ///     1| N A   C D E
  ///     0| N A B C D E
  ///  @endcode
  ///
  ///  and adjust the following:
  ///
  ///    @li level_[].node_ info for levels occupied previously by A now
  ///        refer to N.
  ///
  ///    @li the level_[].offset_ for levels higher than N now become 0
  ///        and N's link_[].width_ member becomes the old offset_ value plus N's
  ///        width.
  ///
  ///  Finally, if the data won't fit in A and also won't fit in a single
  ///  node N, but has overflow components, this routine puts the part that
  ///  will fit into N where it should be then calls the random insert
  ///  routine to jam in the rest after N.

{
  if(last_ == 0)
  {
    push_back(first, last);
    return end();
  }

  ptrdiff_t count = last - first;

  node *n = level_[0].head_;

  ptrdiff_t avail = chunk_size - n->size();

  // handle the simple case where the new data fits in the first node

  if(count <= avail)
  {
    level_[0].head_->insert(first, last, 0);
    size_ += count;

    for(int i=n->levels(); i < active_levels_; ++i)
    {
       level_[i].offset_ += count;
    }

    return iterator(n, count, this);

  }

  // handle the case that a new node must be created

  node *new_node = node::new_node(max_index_levels);

  ++nodes_;

  ptrdiff_t first_node_size = count > chunk_size ? chunk_size : count;

  new_node->insert(first, first + first_node_size, 0);

  size_ += first_node_size;

  int i;

  int extant_levels = new_node->levels();

  if(extant_levels > active_levels_)
    extant_levels = active_levels_;

  for(i=0; i < extant_levels; ++i)
  {
    new_node->link_[i].next_ = level_[i].head_;
    level_[i].head_->link_[i].prev_ = new_node;
    new_node->link_[i].width_ += level_[i].offset_; 

    level_[i].head_ = new_node;
    level_[i].offset_ = 0; // should be zero before, but who knows....
  }

  while(i < active_levels_ )
  {
    level_[i].offset_ += first_node_size;
    ++i;
  }

  while(active_levels_ < new_node->levels())
  {
    level_[active_levels_].offset_ = 0;
    level_[active_levels_].head_   = new_node;
    ++active_levels_;
  }

  // finally, handle the case where more data remains to be
  // inserted after the newly created node

  count -= first_node_size;
  first += first_node_size;

  if(count)
  {
    return insert( iterator(new_node->link_[0].next_, 0, this),
		   first,
		   last
		 );
  }

  return iterator(new_node, new_node->size(), this);

}


template<class T>
typename skiprope<T>::iterator
skiprope<T>::insert(typename skiprope<T>::iterator where,
		    T const *first,
		    T const *last
		   )
//
///   Random inserts are particularly problematic because extant nodes may need to be split
///   and data could theoretically be moved from one node to the other to preserve balance.
///   Here is an example of an insertion that spreads across multiple nodes.  Nodes are limited
///   to 5 objects.
/// 
///   @code
///     11111 22222 33333 44444
///             ^
///             AAAAAAAAAAA (11 A's)
/// 
///   After insertion, we have:
/// 
///     11111 22AAA AAAAA AAA  33333 44444
/// 
///   @endcode
///   Note that we _could_ have merged the trailing 3 A's into the group of 5 33333's but
///   that would convert the operation into an O(N) because you'ld then have to push the trailing
///   3's into the 4's nodes, etc.
/// 
{

  //
  //  handle the easy special cases
  //
  if(where == end())
  {
    push_back(first, last);
    return end();
  }
  else if(where == begin())
  {
    return push_front(first, last);
  }

  node *n         = where.node_;
  ptrdiff_t count = last - first;
  ptrdiff_t avail = chunk_size - n->size();

  if(count <= avail)
  {
    //
    //  In this case, the inserted data will fit in the specifed node
    //  so all we have to do is stick it in and adjust the linkage info.
    //  Here is a worst case scenario:
    //
    //    lvl  Nodes
    //    ---+-------------
    //     3|             G
    //     2|       D     G
    //     1|   B   D   F G
    //     0| A B C D E F G
    //                ^
    //                insert into E
    //
    //  In this case, the link_ info for D must be adjusted for levels
    //  1 and 2, and the level_[] info for the skiprope must be adjusted
    //  for level 3.
    //

    n->insert(first, last, where.offset_);

    size_ += last - first;

    node_info pred[max_index_levels];

    int pred_levels = prednodes(n,pred);

    // pred[].offset_ not set by this call

    int i = n->levels();

    while(i < pred_levels)
    {
	pred[i].node_->link_[i].width_ += count;
	++i;
    }

    while(i < active_levels_)
    {
	level_[i].offset_ += count;
	++i;
    }

    where.offset_ += count;

    return where;
  }
  //
  //  now handle the cases where one or more nodes have to be
  //  injected into the list:
  //
  //      * split where.node_
  //
  //      * tack as much of the inserted data onto the node as
  //        possible
  //
  //      * insert new nodes for data that overflows the start
  //        node
  //
  //  Update linkages as needed
  //


  // SPLIT WHERE.NODE_

  split_node(n, where.offset_);

  avail = chunk_size - n->size();

  if(avail > count)
    avail = count;  // unlikely, but who knows what split_node might do
		    // to n.

  // the following iterator construction uses a special skiprope_iterator()
  // invocation that lets us force the constructor not to search past the
  // empty node we are actually point to.

  insert(iterator(n,where.offset_, this, 0), first, first + avail);

  where = iterator(n, where.offset_ + avail, this);  // will auto wrap to the
						     // next node

  first += avail;
  count -= avail;

  while(count)
  {
    avail = count > chunk_size ? chunk_size : count;

    node *new_node = node::new_node(max_index_levels);

    new_node->insert(first, first + avail, 0);

    insert_node(n, new_node);

    where = iterator(new_node, avail, this);

    first += avail;
    count -= avail;
  }

  return where;

}

template<class T>
void
skiprope<T>::
split_node(typename skiprope<T>::node *n, ptrdiff_t offset)
  //
  ///  Split the specified node into two pieces, starting at the specfied
  ///  offset and update all tables as necessary.  Here are some cases that
  ///  have to be dealt with:
  ///
  ///  @code
  ///  CASE 1:
  ///
  ///    lvl  Nodes                 lvl  Nodes
  ///    ---+------------    ==>>   ---+-------------
  ///     3|           F             3|               F
  ///     2|       D   F             2|           D   F
  ///     1|   B   D E F             1|       B   D E F
  ///     0| A B C D E F             0| a1 a2 B C D E F
  ///        ^                          ^^^^^
  ///
  ///     In this case, A becomes a1 and a2 and both keeps the same level
  ///     as A.  In this case all the level_ info stays the same and only
  ///     the link_'s for a1, a2, and B need to change.
  ///
  ///  CASE 2:
  ///
  ///    lvl  Nodes                 lvl  Nodes
  ///    ---+------------    ==>>   ---+-------------
  ///     3|           F             3|               F
  ///     2|       D   F             2|        c2 D   F
  ///     1|   B   D E F             1|   B    c2 D E F
  ///     0| A B C D E F             0| A B c1 c2 D E F
  ///            ^                          ^^^^^
  ///  @endcode
  ///
  ///    In this case, the second half of the split node has a higher
  ///    level count that the original (ie and now first half).  In this
  ///    case the level_[] info needs to be for levels 2 and higher
  ///    to be adjusted.
  ///
  ///    For example, before the split, level_[2].offset_ was width(A) +
  ///    width(B) + width(C).  After the split, level_[2].offset_ becomes
  ///    width(A)+width(B)+width(C1) and  c2.link_[2].offset_ = width(c2).
  ///
  ///    Before the split, B.link_[1] = {D, width(B) + width(C)};  After
  ///    the split, B.link_[1] = { c2, width(B) + width(c1)}
  //
{
  if(offset >= n->size())
  {
    return;
  }

  // note:  this code doesn't work if n is the last node.  you should
  //        be calling push_back to handle that.


  node_info pred[max_index_levels];

  int pred_levels = prednodes(n, pred); // need nodes and offset info

  //
  //  create a new node to put the split off part in.  Copy the
  //  data to the new node and remove it from the old
  //

  node *new_node = node::new_node(max_index_levels);

  new_node->insert(&(*(n->begin() + offset)), 
                   &(*(n->begin()+n->size())), 
                   0);

  ptrdiff_t deleted_amount = n->size() - offset;

  n->erase(n->begin() + offset, n->end());

  //
  //  adjust the tables to account for the loss
  //
  int i;

  for(i = n->levels(); i < pred_levels; ++i)
  {
    pred[i].node_->link_[i].width_ -= deleted_amount;
  }

  while(i < active_levels_ )
  {
    level_[i].offset_ -= deleted_amount;
    ++i;
  }

  //
  //  connect the split off part up after original owner
  //  in its own node
  //
  size_ -= deleted_amount;

  insert_node(n, new_node);


}

template<class T>
void skiprope<T>::
erase(typename skiprope<T>::iterator location)
{
  node *n = location.node_;

  // this code assumes that location has been adjusted such that it does not
  // point to the end of a node, but has been adjusted to point to the beginning
  // of the next block (iff needed) as is the normal behavior of the iterator
  // constructor

  n->erase(location.offset_, 1);
  size_ -= 1;

  node_info pred[max_index_levels];

  int pred_levels = prednodes(n, pred);

  int i;


  if(n->size() == 0)
  {
    // erasure resulted in a zero length node:  nuke it

    int       n_levels = n->levels();

    for(i=0; i < n_levels; ++i)
    {

      if(n->link_[i].next_)
      {
	n->link_[i].next_->link_[i].prev_ = n->link_[i].prev_;
      }

      if(n->link_[i].prev_)
      {
       
	n->link_[i].prev_->link_[i].next_ = n->link_[i].next_;
        n->link_[i].prev_->link_[i].width_ += n->link_[i].width_;
        // width in 'n' has already had n's size() subtracted out
      }

      if(n == level_[i].head_)
      {
	if(n->link_[i].next_)
	{
	  level_[i].head_ = n->link_[i].next_;
	  level_[i].offset_ += n->link_[i].width_; // n->size is 0, but width may not be
	}
	else
	{
	  level_[i].head_ = 0;
	  level_[i].offset_ = 0;
	  if(i < active_levels_)
	    active_levels_ = i;
	}
      }
    }

    for(; i < pred_levels; ++i)
    {
      // handle the cases where n is shorter than its predecessors
      // specifically handle the levels above n

      if(pred[i].node_->link_[i].next_)
      {
	pred[i].node_->link_[i].width_ -= 1;
      }

    }

    while(i < active_levels_)
    {
      level_[i].offset_ -= 1;
      ++i;
    }

   if(n == last_)
   {
    last_ = n->link_[0].prev_;
   }


    node::discard(n);
    --nodes_;

  }
  else
  {
    // erasure only removed part of the data

    for(i=n->levels(); i < pred_levels; ++i)
    {
      pred[i].node_->link_[i].width_ -= 1;
    }

    while(i < active_levels_)
    {
      level_[i].offset_ -= 1;
      ++i;
    }

  }

}

} // namespace cxxtls

#endif
