//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/dateTools.h>
#include <stdio.h>

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

using namespace std;

namespace cxxtls {
namespace DateTools {


char const *dayAbbrev(int day)
{
   static char const *tab[]=
   {
      "Sun",
      "Mon",
      "Tue",
      "Wed",
      "Thu",
      "Fri",
      "Sat"
   };

   return tab[ day % 7 ];
}

char const *dayFull(int day)
{
   static char const *tab[]=
   {
      "Sunday",
      "Monday",
      "Tuesday",
      "Wednesday",
      "Thursday",
      "Friday",
      "Saturday"
   };

   return tab[ day % 7 ];
}

char const *monthAbbrev(int month)
{
   static char const *tab[]=
   {
      "Jan",
      "Feb",
      "Mar",
      "Apr",
      "May",
      "Jun",
      "Jul",
      "Aug",
      "Sep",
      "Oct",
      "Nov",
      "Dec"
   };

   return tab[ month % 12 ];

}

char const *monthFull(int month)
{
   static char const *tab[]=
   {
      "January",
      "February",
      "March",
      "April",
      "May",
      "June",
      "July",
      "August",
      "September",
      "October",
      "November",
      "December"
   };

   return tab[ month % 12 ];
}



    string
    DateFormatter::
    operator() (struct tm const *tim, time_t secs)
    {
        string rv;

        char buffer[1000]; // for snprintf calls

        string::const_iterator f = format_.begin(),
                               l = format_.end();

        // iterate over the format and populate the output with
        // data specified in the format string.


        while(f != l)
        {
            char c = *f++;

            if(c != '%')
               rv += c;     // normal characters in the format get copied directly to the output
            else
            {
               //
               //  commands in the format string look like this:
               //
               //    %<cmdChar>
               //
               //  Where the <cmdChar> tells you what to do
               //

               if(f == l)
                  rv += '%';  // out of data, nothing to interpret
               else
               switch(c = *f++)
               {
                   case '%':   rv += '%'; break;

                   case 'a':
                   {
                       // abbrev for day of week

                       rv += dayAbbrev(tim->tm_wday);

                   }
                   break;

                   case 'A':
                   {
                       // full day of week

                       rv += dayFull( tim->tm_wday );
                   }
                   break;

                   case 'b':
                   case 'h':
                   {

                       // abbrev of month

                       rv += monthAbbrev( tim->tm_mon );

                   }
                   break;

                   case 'B':
                   {
                       rv += monthFull( tim->tm_mon );

                   }
                   break;

                   case 'c':
                   {
                       // the date and time (e.g., Thu Mar 3 23:05:25 2005).

                       snprintf(buffer, sizeof(buffer), "%s %s %02d %02d:%02d:%02d %04d",
                                dayAbbrev(tim->tm_wday),
                                monthAbbrev(tim->tm_mon),
                                tim->tm_mday,
                                tim->tm_hour,
                                tim->tm_min,
                                tim->tm_sec,
                                tim->tm_year + 1900
                               );

                       rv += buffer;
                   }
                   break;

                   case 'C':
                   {
                      // current century.  Same as %Y except that there is no decade info

                      snprintf(buffer, sizeof(buffer), "%02d",  ( tim->tm_year  + 1900 ) / 100 );

                      rv += buffer;

                   }
                   break;

                   case 'd':
                   {
                      // day of month

                      snprintf(buffer, sizeof(buffer), "%02d",  tim->tm_mday); 

                      rv += buffer;

                   }
                   break;

                   case 'D':
                   {
                      // full day of month mm/dd/yy

                      snprintf(buffer, sizeof(buffer), "%02d/%02d/%02d",  tim->tm_mon+1, tim->tm_mday, (tim->tm_year+1900) % 100); 

                      rv += buffer;

                   }
                   break;

                   case 'e':
                   {
                      // day of month, space padded, not with leading 0

                      snprintf(buffer, sizeof(buffer), "%2d",  tim->tm_mday);    // no 0 in %2d on purpose!

                      rv += buffer;

                   }
                   break;

                   case 'F':
                   {
                       // full date; same as %Y-%m-%d

                      snprintf(buffer, sizeof(buffer), "%04d-%02d-%02d",  tim->tm_year + 1900, tim->tm_mon+1, tim->tm_mday); 

                      rv += buffer;

                   }
                   break;

                   case 'H':
                   {
                      // hour in 24 hour format (00-23)

                      snprintf(buffer, sizeof(buffer), "%02d",  tim->tm_hour); 

                      rv += buffer;

                   }
                   break;

                   case 'I':
                   {
                      // hour in 12 hour format (1-12)

                      int hour = tim->tm_hour % 12;

                      if(hour == 0)
                         hour = 12;

                      snprintf(buffer, sizeof(buffer), "%02d",  hour); 

                      rv += buffer;

                   }
                   break;

                   case 'j':
                   {
                      // julian date, 001-366

                      snprintf(buffer, sizeof(buffer), "%03d",  tim->tm_yday+1); 

                      rv += buffer;

                   }
                   break;

                   case 'k':
                   {
                      // hour in 24 hour format (0-23), space padded

                      snprintf(buffer, sizeof(buffer), "%2d",  tim->tm_hour);   // no zero in the format on purpose

                      rv += buffer;

                   }
                   break;

                   case 'l':
                   {
                      // hour in 12 hour format (1-12), space padded

                      int hour = tim->tm_hour % 12;

                      if(hour == 0)
                         hour = 12;

                      snprintf(buffer, sizeof(buffer), "%2d",  hour);   // no 0 in %0d on purpose

                      rv += buffer;

                   }
                   break;


                   case 'm':
                   {
                      // month, 01-12

                      snprintf(buffer, sizeof(buffer), "%02d",  tim->tm_mon + 1);   

                      rv += buffer;

                   }
                   break;

                   case 'M':
                   {
                      // minute, 00-59

                      snprintf(buffer, sizeof(buffer), "%02d",  tim->tm_min);   

                      rv += buffer;

                   }
                   break;

                   case 'n':
                   {
                      rv += '\n';
                   }
                   break;

                   case 'p':
                   {
                      if(tim->tm_hour < 12)
                         rv += "AM";
                      else
                         rv += "PM";
                   }
                   break;

                   case 'P':
                   {
                      if(tim->tm_hour < 12)
                         rv += "am";
                      else
                         rv += "pm";
                   }
                   break;

                   case 'r':
                   {
                      // locale's 12-hour clock time (e.g., 11:11:04 PM).

                      int hour = tim->tm_hour % 12;

                      if(hour == 0)
                         hour = 12;

                      snprintf(buffer, sizeof(buffer), "%02d:%02d:%02d %s",
                                       hour,
                                       tim->tm_min,
                                       tim->tm_sec,
                                       (tim->tm_hour < 12 ? "AM" : "PM")
                              );

                      rv += buffer;
                   }
                   break;

                   case 'R':
                   {
                       // 24-hour hour and minute; same as %H:%M.

                       snprintf(buffer, sizeof(buffer), "%02d:%02d",
                                tim->tm_hour,
                                tim->tm_min
                               );

                       rv += buffer;

                   }
                   break;

                   case 's':
                   {
                       // utcs seconds

                       snprintf(buffer, sizeof(buffer), "%ld", (long)(secs));

                       rv += buffer;

                   }
                   break;

                   case 'S':
                   {
                       // seconds relative to the current minute

                       snprintf(buffer, sizeof(buffer), "%02d",
                                tim->tm_sec
                               );

                       rv += buffer;

                   }
                   break;

                   case 't':
                   {
                       rv += '\t';  // tab
                   }
                   break;

                   case 'T':
                   {
                       // time; same as %H:%M:%S.

                       snprintf(buffer, sizeof(buffer), "%02d:%02d:%02d",
                                tim->tm_hour,
                                tim->tm_min,
                                tim->tm_sec
                               );

                       rv += buffer;

                   }
                   break;

                   case 'u':
                   {
                       // day of the week (1-7), 1 is monday.

                       int day = tim->tm_wday % 7;

                       if(day == 0)
                          day = 7;

                       snprintf(buffer, sizeof(buffer), "%d",
                                day
                               );

                       rv += buffer;

                   }
                   break;

                   case 'w':
                   {
                      // day of the week (0-6), 0 is sunday

                       snprintf(buffer, sizeof(buffer), "%d",
                                tim->tm_wday
                               );

                       rv += buffer;

                   }
                   break;

                   case 'x':
                   {
                     // locale's date representation (e.g., 12/31/99).

                       snprintf(buffer, sizeof(buffer), "%02d/%02d/%02d",
                                tim->tm_mon + 1,
                                tim->tm_mday,
                                (tim->tm_year + 1900) % 100
                               );

                       rv += buffer;

                   }
                   break;


                   case 'X':
                   {
                       // locale's time representation (e.g., 23:13:48).

                       snprintf(buffer, sizeof(buffer), "%02d:%02d:%02d",
                                tim->tm_hour,
                                tim->tm_min,
                                tim->tm_sec
                               );

                       rv += buffer;

                   }
                   break;

                   case 'y':
                   {
                       // last 2 digits of year

                       snprintf(buffer, sizeof(buffer), "%02d",
                                (tim->tm_year + 1900) % 100
                               );

                       rv += buffer;

                   }
                   break;

                   case 'Y':
                   {
                       // 4 digit year

                       snprintf(buffer, sizeof(buffer), "%04d",
                                (tim->tm_year + 1900)
                               );

                       rv += buffer;

                   }
                   break;



               } // interpret the command character following the %
            }

        } // iterate over format

        return rv;

    }  // DateFormatter::operator()
}
}
