#ifdef __linux__
//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/gdbcontroller.h>
#include <cxxtls/fmtd.h>
#include <cxxtls/varobj.h>
#include <cxxtls/file.h>
#include <cxxtls/strtool.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/simple_regex.h>

#include <signal.h>
#include <iostream>
#include <fstream>
#include <fcntl.h>
#include <time.h>
#include <memory.h>
#include <time.h>
#include <errno.h>

using namespace std;

//=============================================================================
//=============================================================================
//=============================================================================
namespace cxxtls
{


void* gdbControllerReader(void *); // The thread that reads gdb output.  
                                   // Pass &GdbController
                                   // on thread create.

void* gdbControllerTgtReader(void *); // The thread that reads gdb output.  
                                      // Pass &GdbController
                                      // on thread create.



#define DIAGNOSTIC_LOG_BLOCK     if(debugLogging_)   


#define GDBLOG(msg)         { ScopedMutexLock lock(logMut_);    (*gdbLog_)    << msg   << endl;             }
#define GDBLOG_NONL(msg)    { ScopedMutexLock lock(logMut_);    (*gdbLog_)    << msg;  gdbLog_->flush();    }
#define CTRLOG_NONL(c,msg)  { ScopedMutexLock lock(c->logMut_); (*c->gdbLog_) << msg;  c->gdbLog_->flush();    }

//=============================================================================
//=============================================================================
//=============================================================================

GdbController::
GdbController(bool diaglog, std::string const &logfileName)
:  gdbRdr_(0)
,  tgtRdr_(0)
,  pid_(0)
,  state_(GdbController::notStarted)
,  reason_("initally created")
,  chgCnt_(0)
,  stkCnt_(0)
,  varCnt_(0)
,  expCnt_(0)
,  expWat_(0)
,  stopping_(false)
,  debugLogging_(diaglog)
,  gdbLog_(new std::ofstream(logfileName.c_str()))
,  async_("&")
,  eChgCnt_(0)
,  hasSymbols_(true)
,  curDirCnt_(0)
,  doneErrorCnt_(0)
,  doneErrorScooping_(0)
,  firstSetArgs_(true)
,  controllerStarted_(false)
{
}


//=============================================================================
//=============================================================================
//=============================================================================
GdbController::
~GdbController()
{
  if(gdb_.get())
     stopGdb();
}

//=============================================================================
//=============================================================================
//=============================================================================
static void shortNap()
{
     struct timespec delay, rv;

     delay.tv_sec = 0;
     delay.tv_nsec=  100 * 1000 * 1000;   // 100,000,000 nanoseconds or 1/10 of a second

     nanosleep(&delay, &rv);
}

//=============================================================================
//=============================================================================
//=============================================================================
void* gdbControllerReader(void *p)  // The thread that reads gdb output.  
                                    // Pass &GdbController
                                    // on thread create.
{
    GdbController *controller = reinterpret_cast<GdbController*>(p);

    FILE*  gdbOut = controller->gdb_->process_stdout();

    char buffer[20000];

    controller->incrementReaderThreadCount();

    while( fgets(buffer, sizeof(buffer), gdbOut) )
    {
        CTRLOG_NONL(controller,buffer);  // buffer has \n already

        controller->interpretGdbOutput(buffer);
    }

    controller->setState(GdbController::gdbExiting, "end of input on gdb stream");

    controller->decrementReaderThreadCount();

    return 0;
}

//=============================================================================
//=============================================================================
//=============================================================================

void* gdbControllerTgtReader(void *p) // The thread that reads gdb output.  
                                      // Pass &GdbController
                                      // on thread create.
{
    GdbController *controller = reinterpret_cast<GdbController*>(p);

try
{

    int fd = controller->pty_->masterFD();

    GdbController::TargetOutput &output = controller->targetOutput();

    std::string currentLine;

    controller->incrementReaderThreadCount();


    for(;;)
    {
       char c;

       int rc = read(fd, &c, 1);

       ScopedMutexLock lock(controller->stopping_mutex_);

       if(controller->stopping_)
            break;

       if(rc == 0)
       {
           shortNap();


           continue;
       }
       else
       if(rc < 0)
       {
           break;
       }
       else
       {
          output.push(c);

          currentLine.push_back(c);

          if(c == '\n')
          {
             // flush current line to the log file

             CTRLOG_NONL(controller, "<>" << currentLine);

             currentLine.resize(0);

          }

       }

    }

    controller->decrementReaderThreadCount();


    return 0;

}
catch(...)
   {
    controller->decrementReaderThreadCount();

       return 0;
   }
}
//=============================================================================
//=============================================================================
//=============================================================================

void 
GdbController::
printAndDeleteTargetOutput()
{
    std::deque<char> to;

    targetOutput().trySwapAll(to);

    DIAGNOSTIC_LOG_BLOCK
    {
        CXXTLS_COUT_LOCK(var);  // lock cout so we can print a bunch of crap to it
        
        CXXTLS_FOREACH(char c, to)
        {
           cout << c;
        }
    }

}

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
run(int pid)
{

    if(pid)
    {
       // NOT ACTIVE.  Do NOT turn off async async_ = ""; // turn off async operations

       pid_ = pid;

    }

    ChildState s = state();
    
    if(   s == notStarted 
       || s == terminated
       || s == runnable
      )
    {
        if(exe_.empty())
        {
            ScopedMutexLock lock(mutex_);
            reason_ = "executable not specified";
            return false;
        }

        if(s == notStarted)
        {
           if(!startGdb())
           {
              ScopedMutexLock lock(mutex_);
              reason_ = "Unable to run gdb";
              return false;
           }
        }

        s = state();

        if( s != runnable && s != terminated )
        {
            return false;
        }

        {
            ScopedMutexLock lock(controllerStartMutex_);

            if(!controllerStarted_)
            {
                gdbRdr_ = createThread(gdbControllerReader, this);
                controllerStarted_ = true;
            }
        }

        pty_.reset(0); // kill any previous terminal if one existed

        if(tgtRdr_ != 0)
        {

            joinThread(tgtRdr_);

            tgtRdr_ = 0;
        }

        pty_.reset(new PseudoTerminal);


        if(pty_->open(false) != PseudoTerminal::good)
        {
           ScopedMutexLock lock(mutex_);
           reason_ = "Unable to create pty for child to run int";
           return false;
        }

        sendArgsAndSetup();

        if(!cdForRun_.empty())
        {
           setcwd(cdForRun_);
           cdForRun_ = "";
        }

        setState(running, "ok to run");

        string slaveTty = pty_->slaveTerminalName();

        string command = string("tty ") + slaveTty;

        if(pid == 0)
            sendCommand(command);

//        if(pid == 0)
            sendCommand("-gdb-set target-async 1");

        size_t cnt = stateChangeCount();

        if(pid)
        {
            char buffer[256];

            snprintf(buffer, sizeof(buffer), "-target-attach %d", pid);

            sendCommand(buffer);  // perform attachment and cause the target to stop


            bool error = 
            awaitStateChange(cnt,2);  // wait up to 2 seconds for any
                                      // change to the state_ variable.  We should be able to halt a running
                                      // process with in two seconds

            if(error)
            {
               setState(notStarted, "unable to attach");

               // do not start the reader thread if we can't attach

               return false;
            }


        }
        else
        {
            {
               ScopedMutexLock lock(eMutex_);
               eaccum_.resize(0);
               evalues_.clear();
        
            }

            sendCommand(std::string("run ") + async_);  // & for run in background -- so I can send an interrupt command to stop it

            awaitStateChange(cnt);   // wait up to 5 seconds for any
                                     // change to the state_ variable
        }
        

        tgtRdr_ = createThread(gdbControllerTgtReader, this);

        return true;

    }

    return false;
}
//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
sendCommand(std::string const &cmd)
{
   GDBLOG(cmd);

   if(!gdb_.get())
     {
       CXXTLS_DEBUG("SendCommand:  ERROR gdb_ IS NULL for command" << cmd);
     return;  // nothing to do
     }

   FILE* f = gdb_->process_stdin();

   fputs(cmd.c_str(), f);

   DIAGNOSTIC_LOG_BLOCK
   {
       CXXTLS_DEBUG("SendCommand:  " << cmd);
   }


   fputs("\n", f);

}

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
setExecutable(std::string const &e)
{
   FileName name(e);

   if(!name.is_executable())
   {
       if(!name.exists())
       {
          FileName fullpath = FileName::find_executable(e);
       
          if(fullpath.is_executable())
          {
             name = fullpath; // found executable in search path
          }
       }
       else
          return false; // original exec file exists but is not executable
   }

   ScopedMutexLock lock(mutex_);

   dirty_=true;


   exe_ = name;

   return true;
}

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
setArgs(std::string const &a)
{
   ScopedMutexLock lock(mutex_);

   args_ = a;

   dirty_ = true;

   return true;
}

//=============================================================================
//=============================================================================
//=============================================================================
GdbController::ChildState
GdbController::
state()
{
   ScopedMutexLock lock(mutex_);

   return state_;

} // namespace cxxtls


//=============================================================================
//=============================================================================
//=============================================================================
char const *GdbController::stateName_[]=
{
    "notStarted",   
    "runnable", 
    "running",      
    "stopped",      
    "terminated",   
    "unstartable",  
    "gdbExiting",   
};

//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
stateName(GdbController::ChildState i) 
{
     if(size_t(i) < sizeof(stateName_) / sizeof(char*))
     {
         return stateName_[i];
     }

     char buffer[40];

     snprintf(buffer, sizeof(buffer), "unknown[%d]", i);

     return buffer;
}

//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
addMonitoredVariable(std::string const &e)
{
   string rv = nextId('D'); // this does a temp mutex lock, so it must be above 
                            // scoped lock guarding the rest of the data

   ScopedMutexLock lock(mutex_);

   sendCommand( string("-var-create ") + rv + string(" * ") + e);

   variables_.insert( VariableValues::value_type(rv,VariableInfo()) );

   VariableValues::iterator where = variables_.find(rv);

   if(where == variables_.end())
   {
        DIAGNOSTIC_LOG_BLOCK
        {
            CXXTLS_DEBUG("Error:  insert into map failed?  Not possible");
        }
   }
   else
   {
      where->second.expr_ = e;
      where->second.id_   = rv;
   }


   pndVars_.push_back(rv);

   return rv;
}

//=============================================================================
//=============================================================================
//=============================================================================
void 
GdbController::
removeMonitoredVariable(std::string const &e)
{
   ScopedMutexLock lock(mutex_);

   VariableValues::iterator where = variables_.find(e);

   if(where == variables_.end())
   {
      DIAGNOSTIC_LOG_BLOCK
      {
          CXXTLS_DEBUG("Error:  undefined display variable name '" << e << "'");
      }
   }
   else
   {
      variables_.erase(where);
      sendCommand( string("-var-delete ") + e);
   }
}


//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
exceptionBreakpoints(bool catchPoint, bool throwPoint)
{
   if(catchPoint)
       sendCommand("catch catch");

   if(throwPoint)
       sendCommand("catch throw");


   sendCommand("-break-list");  // to make sure the bp list is displayed
                                // correctly

   sleep(1); // really, this should have mutex change counter logic.
}


//=============================================================================
//=============================================================================
//=============================================================================
void 
GdbController::
setBreakpointCondition(std::string const &bpID, std::string const &bpCond)
{
    static std::string space(" ");

    sendCommand(std::string("-break-condition ") + bpID + space + bpCond);
    sendCommand("-break-list");
    sleep(1);  // should really use a mutex locked change count, but this is
               // ok till then

}

//=============================================================================
//=============================================================================
//=============================================================================
bool 
GdbController::
getBreakpointCondition(std::string const &bpID, std::string &bpCond)
{
    Breakpoints::iterator where = breakpoints_.find(bpID);

    if(where == breakpoints_.end())
       return false;

    VarObj* cond = where->second->findNestedChild("cond");

    if(cond)
      {   
         bpCond = cond->value();
      }

    return true;
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
watchLocation(std::string const &frame, std::string const &expression, char const *access)
{


    if(state() == stopped)
    {
        selectStackFrame(frame);

        std::string watchCommand("-break-watch ");
                    watchCommand += access;
                    watchCommand += ' ';
                    watchCommand += expression;

        sendCommand(watchCommand);
        sendCommand("-break-list");

        sleep(1); // should really be some sort of check for a change to the breakpoint list



    }

}

//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
getMostRecentBreakpoint()
{
   ScopedMutexLock lock(mutex_);

   return mrBreakpoint_;
}



//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
addBreakpoint(std::string const &e)
{

    if(state() != stopped)
    {
       ScopedMutexLock lock(mutex_);
       pndBP_.push_back(e);
    }
    else
    {
        size_t startCount = clearDoneErrorMessage();

        sendCommand( string("-break-insert ") + e );

        return awaitDoneErrorChange(startCount);   // wait for interpretGdbOutput to notice the result

    }
    // Note that this function triggers the creation of a response from
    // gdb of the form ^done,bkpt={...}.  
    //
    // See interpretGdbOutput() to see how this response gets stored in the
    // variable, breakpoints_.
    return "";
}

//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
addBreakpoint(std::string const &file, size_t line)
{
    static std::string colon(":");

    std::string e = file + colon + std::string(fmtd<>(line));

    if(state() != stopped)
    {
       ScopedMutexLock lock(mutex_);
       pndBP_.push_back(e);
    }
    else
    {
        size_t startCount = clearDoneErrorMessage();

        sendCommand( string("-break-insert ") + e );

        return awaitDoneErrorChange(startCount);

    }

    // Note that this function triggers the creation of a response from
    // gdb of the form ^,bkpt={...}.  
    //
    // See interpretGdbOutput() to see how this response gets stored in the
    // variable, breakpoints_.
    return "";
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
enableBreakpoint(std::string const &name, bool enabled)
{

    if(enabled)
        sendCommand( string("-break-enable ") + name );
    else
        sendCommand( string("-break-disable ") + name );

    sendCommand("-break-list");
    sleep(1);



}



//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
stepOver()
{
    if(state() == stopped)
    {

       {
          ScopedMutexLock lock(eMutex_);
          eaccum_.resize(0);
          evalues_.clear();
       
       }

       size_t cnt = stateChangeCount();

       sendCommand(std::string("next ") + async_);  // & for run in background

       return !awaitStateChange(cnt);   // wait up to 5 seconds for any
                                        // change to the state_ variable
                                        // return true if no state change
                                        // timeout has occurred
    }

    return false;
}

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
stepOut()
{
    if(state() == stopped)
    {
       size_t cnt = stateChangeCount();

       {
          ScopedMutexLock lock(eMutex_);
          eaccum_.resize(0);
          evalues_.clear();
       
       }

       sendCommand(std::string("finish ") + async_);  // & for run in background

       return !awaitStateChange(cnt);   // wait up to 5 seconds for any
                                        // change to the state_ variable
                                        // return true if no state change
                                        // timeout has occurred
       
    }

    return false;
}


//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
halt()
{
    if(state() == running)
    {
       size_t cnt = stateChangeCount();

       // sendCommand(std::string("interrupt") + async_); 

       if(pid_)
       {
          GDBLOG("Sending SIGINT to pid(" << pid_ << ")");

          ::kill(pid_, SIGINT);
       }
       else
       {
           GDBLOG("Cannot stop target, pid is unknown");

           return true;
       }

       if(awaitStateChange(cnt))     
       {
          return true;
       }

       sendCommand("-stack-list-frames 0 40");
                                     
                                     
                                     
       
    }

    return false;
}


//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
stepInto()
{
    if(state() == stopped)
    {
       size_t cnt = stateChangeCount();

       {
          ScopedMutexLock lock(eMutex_);
          eaccum_.resize(0);
          evalues_.clear();
       
       }

       sendCommand(std::string("step ") + async_); // & for run in background
       return !awaitStateChange(cnt);   // wait up to 5 seconds for any
                                        // change to the state_ variable
                                        // return true if no state change
                                        // timeout has occurred
       
    }

    return false;
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
setState(ChildState s, std::string const &reason)
{
   ScopedMutexLock lock(mutex_);

   state_ = s;
   reason_= reason;
   ++chgCnt_;   // increment count of state changes

} 

//=============================================================================
//=============================================================================
//=============================================================================
size_t 
GdbController::
stateChangeCount() 
{
    ScopedMutexLock lock(mutex_);

    return chgCnt_;
    
}

//=============================================================================
//=============================================================================
//=============================================================================
size_t 
GdbController::
argsChangeCount() 
{
    ScopedMutexLock lock(stkMut_);

    return stkCnt_;
    
}

//=============================================================================
//=============================================================================
//=============================================================================
size_t 
GdbController::
variablesChangeCount() 
{
    ScopedMutexLock lock(stkMut_);

    return varCnt_;
    
}


//=============================================================================
//=============================================================================
//=============================================================================

bool
GdbController::
awaitExpressionChange(size_t startChangeCount, int seconds)
{
   int maxSeconds = seconds * 10;  // 10 shortNaps per second

   for(int i = 0; i < maxSeconds; ++i)  
   {
      if(startChangeCount == expressionChangeCount())
         shortNap();  // sleep 0.1 seconds
      else
         return false;
   }

   GDBLOG("Time out waiting for GDB state change");

   return true; // if we get here, a time out has occurred
}

//=============================================================================
//=============================================================================
//=============================================================================

bool
GdbController::
awaitStateChange(size_t startChangeCount, int seconds)
{
   int maxSeconds = seconds * 10;  // 10 shortNaps per second

   for(int i = 0; i < maxSeconds; ++i)  
   {
      if(startChangeCount == stateChangeCount())
         shortNap();  // sleep 0.1 seconds
      else
         return false;
   }

   GDBLOG("Time out waiting for GDB state change");


   return true; // if we get here, a time out has occurred
}

//=============================================================================
//=============================================================================
//=============================================================================

bool
GdbController::
VariableInfo::
awaitChange(size_t startChangeCount, int seconds)
{
   int maxSeconds = seconds * 10;  // 10 shortNaps per second

   for(int i = 0; i < maxSeconds; ++i)  
   {
      if(startChangeCount == changeCount())
         shortNap();  // sleep 0.1 seconds
      else
         return false;
   }

   return true; // if we get here, a time out has occurred
}

//=============================================================================
//=============================================================================
//=============================================================================

bool
GdbController::
awaitStackParmChange(size_t startChangeCount, int seconds)
{
   int maxSeconds = seconds * 10;  // 10 shortNaps per second

   for(int i = 0; i < maxSeconds; ++i)  
   {
      if(startChangeCount == argsChangeCount())
         shortNap();  // sleep 0.1 seconds
      else
         return false;
   }

   GDBLOG("Time out gdb to produce stack frame parameters");

   return true; // if we get here, a time out has occurred
}

//=============================================================================
//=============================================================================
//=============================================================================

bool
GdbController::
awaitVariableChange(size_t startChangeCount, int seconds)
{
   int maxSeconds = seconds * 10;  // 10 shortNaps per second

   for(int i = 0; i < maxSeconds; ++i)  
   {
      if(startChangeCount == variablesChangeCount())
         shortNap();  // sleep 0.1 seconds
      else
         return false;
   }


   GDBLOG("Time out gdb to produce stack frame variables");


   return true; // if we get here, a time out has occurred
}



//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
resume()
{
    ChildState s = state();

    if(s != stopped)
       return false;


    {
       ScopedMutexLock lock(eMutex_);
       eaccum_.resize(0);
       evalues_.clear();

    }


    size_t cnt = stateChangeCount();

    sendCommand(std::string("continue ") + async_);

    return !awaitStateChange(cnt);   // wait up to 5 seconds for any
                                     // change to the state_ variable
                                     // return true if no state change
                                     // timeout has occurred
    

} 
//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
startGdb()
{
    if(state() != notStarted)
    {
       return false;
    }

    if(gdb_.get() == 0)
    {
        std::string commandLine = "gdb -q --interpreter=mi";


        gdb_.reset(   ControlledProcess::start(commandLine,
                                               ControlledProcess::LINE_BUFFERED, 
                                               ControlledProcess::CHARACTER_BUFFERED
                                              )
                  );

        if(gdb_->status() != 0)
        {
             return false;
        }

        // at this point, gdb does not know to use the pseudo terminal.  A
        // sendCommand("tty <name>") will be used when the running starts

        setState(runnable, "starting gdb");
        reason_ = "gdb is running";

    }

    return true;

} 

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
kill()
{
    if(state() == running)
    {
         
         size_t cnt = stateChangeCount();
         
         sendCommand("kill ");  // forces the target into the terminated state
         
         return !awaitStateChange(cnt);   // wait up to 5 seconds for any
                                          // change to the state_ variable
                                          // return true if no state change
                                          // timeout has occurred
         
    }

    sendCommand("kill");  
    sleep(1);  // should really wait for "thread-group-exited"

    return false;

}
//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
saveState()
{
    if(firstSetArgs_ || state() == notStarted)
        return false; 
                // user has not yet run the debugger and we depend on some
                // initialization logic to populate the variables saved by
                // this function.   So basically, if you don't run, you can't
                // save.


    std::string filename = saveFile();

    if(filename.empty())
       return false; // nothing to do

    if(saveFile_.empty())
       saveFile_ = filename;

    {
        ofstream bpOutput(saveFile_.c_str());

        if(bpOutput.good())
        {
            bpOutput << "{\n";

            bpOutput << "    dir="  << VarObj::encodeString(curDir_) << ',' << endl;
            bpOutput << "    args=" << VarObj::encodeString(args_)   << ',' << endl;
            bpOutput << "    exeFile=" << VarObj::encodeString(exe_) << ',' << endl;


            bpOutput << "    breakpoints=[\n";


            CXXTLS_FOREACH(Breakpoints::value_type const &cur, breakpoints_)
            {

                VarObj *fullname = cur.second->findChild("fullname");
                VarObj *line     = cur.second->findChild("line");

                if(fullname && line)
                {
                    bpOutput << "        {fullname=\"" << fullname->value() << "\"," << "line=" << line->value() << '}';
                }



                bpOutput << '\n';
            }


            bpOutput << "    ]\n";

            bpOutput << "}\n";


            return false; // no errors
        }

    }

    return true; // could not open the save file, I guess.
}
//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
stopGdb()
{
    saveState();

    sendCommand("quit");  // should kill gdb and the debugged child, which should
                          //terminate the two threads

    shortNap(); // give the quit command some time to work
    shortNap();

    gdb_.reset(0);  // kill off gdb if it isn't already dead.

    if(pty_.get())
    {
       ScopedMutexLock lock(stopping_mutex_);
       stopping_ = true;

       write(pty_->masterFD(), "\n", 1);

       shortNap();
       shortNap();

       pty_->close();    // make sure tgtRdr_ terminates even gdb never started

       shortNap();

       tgtRdr_ = 0;

    }

    // When we get here, hopefully gdb will have terminated and the GdbControllerReader() will
    // detect end of of file on the output file from gdb


    joinThread(gdbRdr_);
    joinThread(tgtRdr_);
} 

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
interpretGdbOutput(char const*line)
{
    // Every line of output from gdb/MI mode gets run through this function.
    // Lines are formatted like this:
    //
    //    <c><type>[ , <VarObjList> ]
    //
    // Where <c> is one of
    //
    //   *
    //   ^
    //   =
    //
    // Note that debugged target program output does not come through here
    // because when gdb is invoked, the -tty option is used to instruct it
    // to launch the target in a separate pseudo terminal.
    //
    // This function interprets the <type> string to determine what to do.
    // See below.
    //
    // Also, do a google search for gdb/mi for a description of the gdb
    // interpreter mode.

    DIAGNOSTIC_LOG_BLOCK
    {
        CXXTLS_DEBUGNONL(line); // alread has \n on it
    }

    ChildState curState = state();

    static string interestingLineTypes("*=^");

    if(interestingLineTypes.find_first_of(line[0]) < interestingLineTypes.size())
    {
        string tmp(line);

        if(tmp[ tmp.size() - 1] == '\n')
        {
            tmp.erase(tmp.size() -1);  // remote trailing \n
        }

        if(tmp.empty())
            return;

        if(tmp.size() >= 5 && tmp[0] == '^' && tmp[1] == 'd' && tmp[2] == 'o' && tmp[3] == 'n')
        {
                 GDBLOG("Found ^done");

                 bool doit = false;

                 {

                     ScopedMutexLock lock(eMutex_);

                     static SimpleRegex digitColon("[0-9]\\+:");

                     GDBLOG("eaccum_ ==>" << eaccum_);

                     int matchCount = digitColon(eaccum_);

                     if(matchCount)
                     {
                        doit = true;

                        SimpleRegex::match const * matches = digitColon.matches();

                        eaccum_.erase(0, matches->offset);

                        GDBLOG("Updated eaccum==>" << eaccum_);


                     }

                 
                 
                 }

                 if(doit)
                     recordDisplayExpressionResults();

        }



        // most lines we are interested in have the [, <VarObjectList. ] attatched
        // so check to make sure they exist.  The one exception is the *stopped
        // type which may not have a trailing var object list.


        size_t comma = tmp.find_first_of(',');

        if(comma >= tmp.size())
        {

           static string completeStop("*stopped");
           static string simpleDone("^done");

           if(completeStop == tmp)
           {
               setState(terminated, "gdb signal terminated");
           }
           else
             if(simpleDone == tmp)
               {



                 incDoneErrorCount(); // ^done with nothing after it
               }


           return;
        }

        //////////////////////////////////////////////////////////////////////////////////
        // now interpret gdb message types
        //////////////////////////////////////////////////////////////////////////////////

        string firstWord(tmp.substr(0,comma));

        if(firstWord == "*stopped")
        {
            //////////////////////////////////////////
            // the target process has stopped running
            //////////////////////////////////////////

            char const *f = tmp.data() + comma+1;
            char const *l = tmp.data() + tmp.size();

            string  reason("gdb detected the target stopping");

            ChildState newState=stopped;


            VarObj::ParseResult  r = VarObj::parseList(f,l);

            if(r.rv_ == 0)
            {
                DIAGNOSTIC_LOG_BLOCK
                {
                    CXXTLS_DEBUG("Cannot parse gdb output:  " << tmp << "\n" << r.error_);
                }

            }
            else
            {
                adoptStopInfo(r.rv_);

                CompositeVarObj *info =  dynamic_cast<CompositeVarObj*>(r.rv_);

                if(!info)
                {
                    DIAGNOSTIC_LOG_BLOCK
                    {
                         CXXTLS_DEBUG("Did not find file/line in gdb output" << tmp);
                    }
                }
                else
                {
                   VarObj *filePtr = info->findNestedChild("frame/fullname");
                   VarObj *linePtr = info->findNestedChild("frame/line");
                   VarObj *reasonP = info->findChild("reason");

                   if(!filePtr)
                   {
                      filePtr = info->findNestedChild("frame/file");
                   }

                   string  file("unknown");
                   string  line("unknown");

                   if(filePtr) 
                      file = filePtr->value();

                   if(linePtr)
                      line = linePtr->value();

                   if(reasonP)
                      reason = reasonP->value();

                   DIAGNOSTIC_LOG_BLOCK
                   {
                        CXXTLS_LOG("Stopped at file, " << file << ", line " << line << ", because of " << reason);
                   }

                   if(    (reason == "exited-normally")
                      ||  (reason == "exited-signalled")
                      ||  (reason == "exited")
                     )
                      newState = terminated;


                }

            }


            if(    (newState != curState) 
               &&  (newState != terminated) 
              )
            {
                sendCommand("-thread-info");

                {
                    ScopedMutexLock lock(mutex_);

                    if(!variables_.empty())
                        sendCommand("-var-update *");
                }

                {
                    ScopedMutexLock lock(mutex_);

                    sendCommand("-stack-list-frames 0 40");
                    shortNap(); // give the stack a short amount of time to return
                }


            }

            setState(newState, reason);

        }
        else
        if(firstWord == "*running")
        {
            //////////////////////////////////////////
            // The target process has started running
            //////////////////////////////////////////

            setState(running, "gdb signaled that it has started / resumed target execution");

            if(!pndVars_.empty())
            {
                DIAGNOSTIC_LOG_BLOCK
                {
                    CXXTLS_LOG("Error:  gdb seems to have run the debugged child target process"
                               << "\n        but GdbController has not yet received values"
                               << "\n        for the following display variables:"
                              );
                    
                    CXXTLS_FOREACH(std::string const &cur, pndVars_)
                    {
                        CXXTLS_LOG("       '" << cur << "'" );
                    }
                }

                pndVars_.clear();
            
            }

        }
        else
        if(firstWord == "^done")
        {
            ///////////////////////////////////////////
            // Gdb has finished executing a data query
            // command that we have issued to it.
            ///////////////////////////////////////////

            // we have found ^done,....
            //
            // Note:  ^done can occur without any trailing ,.... and we will not
            //        get here under that circumstance

            char const *f = tmp.data() + comma+1;
            char const *l = tmp.data() + tmp.size();

            VarObj::ParseResult  rval = VarObj::parseList(f,l);

            if(rval.rv_ == 0)
            {
                DIAGNOSTIC_LOG_BLOCK
                {
                     CXXTLS_DEBUG("Error:  cannot parse:  " << line << "Error: " << rval.error_);
                }
            }
            else
            {
                CompositeVarObj* c = dynamic_cast<CompositeVarObj*>(rval.rv_);

                if(!c)
                {
                     DIAGNOSTIC_LOG_BLOCK
                     {
                          CXXTLS_DEBUGNONL("Error: expected list of threads info object, but found:  " << line); 
                     }
                }
                else
                {
                    ////////////////////////////////////////////////////////////////////////
                    // Done commands fit into these broad outlines:
                    //
                    //   ^done,value=
                    //   ^done,threads=
                    //   ^done,changelist=
                    //   ^done,numchild="3",children=
                    //   ^done,stack=
                    //   ^done,stack-args=
                    //   ^done,variables=
                    //
                    // Further, the value= lines may or may not have other data beside the
                    // value.  For example, the name of the variable for which the value
                    // applies may or may not be supplied.  When it is not suppplied, the
                    // assumption is that the queue, pndVars_, will supply the name.
                    // When it is there, a test is made to validate that pndVars_.front()
                    // matches the actual output.
                    ////////////////////////////////////////////////////////////////////////

                    std::auto_ptr<VarObj> threads    (c->release("threads"));     // find and extract the threads child
                    std::auto_ptr<VarObj> changelist (c->release("changelist"));  // find and extract the changelist child
                    std::auto_ptr<VarObj> bkpt       (c->release("bkpt"));        // find and extract the newly added breakpoint child
                    std::auto_ptr<VarObj> bpTable    (c->release("BreakpointTable"));  // find the list of breakpoints
                    std::auto_ptr<VarObj> stack      (c->release("stack"));       // find and extract the newly added stack child
                    std::auto_ptr<VarObj> args       (c->release("stack-args"));  // find and extract the newly added stack-args child
                    std::auto_ptr<VarObj> variables  (c->release("variables"));   // find and extract the newly added variables child

                    VarObj*       done_value  = c->findChild("value");  // just find the value child
                    ArrayVarObj*  children    = dynamic_cast<ArrayVarObj*>(c->findChild("children"));

                    if(threads.get())
                    {
                        // handle a non-empty list of threads

                        ArrayVarObj *array = dynamic_cast<ArrayVarObj*>(threads.get());

                        if(array)
                        {
                            // we did in fact get an array of threads

                            ScopedMutexLock lock(mutex_);

                            
                            threads_.reset( threads.release() );
                            
                            if(!array->empty())
                            {
                                
                                VarObj* curThread = c->findChild("current-thread-id");
                                
                                if(curThread)
                                {
                                    curThread_ = curThread->value();
                                }
                                else
                                {
                                    DIAGNOSTIC_LOG_BLOCK
                                    {
                                         CXXTLS_DEBUGNONL("Error:  expected to find cur-thread-id in " << line);
                                    }
                                }
                            }
                            
                        }
                        else
                        {
                            DIAGNOSTIC_LOG_BLOCK
                            {
                                CXXTLS_DEBUGNONL("Error:  expected an array of threads but found something else: " << line);
                            }
                        }

                    } // threads=
    
                    if(changelist.get())
                    {
                        // handle a non-empty list of changes

                        ArrayVarObj *changes = dynamic_cast<ArrayVarObj*>(changelist.get());


                        if(changes)
                        {
                            size_t count = changes->size();

                            ScopedMutexLock lock(mutex_);

                            changelist_.reset( changelist.release() );
                            
                            if(curState != terminated)
                            {
                                for(size_t i = 0; i < count; ++i)
                                {
                                    CompositeVarObj *var = dynamic_cast<CompositeVarObj*>( (*changes)[i] );

                                    if(var)
                                    {
                                        // we have the attributes for the current variable

                                        VarObj* name = var->findChild("name");

                                        if(name)
                                        {
                                            // we can in fact find the variable of interest

                                            VarObj* inScope = var->findChild("in_scope");

                                            if(inScope && inScope->value() == "true")
                                            {
                                               // mark the variable as being in scope and
                                               // trigger a read of its value

                                               VariableValues::iterator where = variables_.find(name->value());

                                               if(where != variables_.end())
                                               {
                                                   where->second.inScope_ = true;
                                               }

                                                pndVars_.push_back(name->value());
                                                sendCommand( string("-var-evaluate-expression ") + name->value() );
                                            }
                                            else
                                            {
                                               // Mark the variable as being no longer in scope.

                                               VariableValues::iterator where = variables_.find(name->value());

                                               if(where != variables_.end())
                                               {
                                                   where->second.inScope_ = false;
                                               }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else
                        {
                            DIAGNOSTIC_LOG_BLOCK
                            {
                                CXXTLS_DEBUGNONL("Error:  expected an array of changed variables but found something else: " << line);
                            }
                        }
                    } // changelist=

                    if(done_value)
                    {  // ^done,value=...

                        ScopedMutexLock lock(mutex_);

                        string value = done_value->value();

                        VarObj* specifiedName = c->findChild("name");

                        // we are always expecting a stack, pndVars_, to have the name of the item
                        // whose value we are reading to be at the top.  However, sometimes the name
                        // gets included in the gdb output and sometimes not.
                        //
                        // we always pop off the name from the stack, and if we can we check it
                        // against gdb output.

                        if(!pndVars_.empty() || specifiedName)
                        {
                            string name;

                            if(!pndVars_.empty())
                            {
                                 name = pndVars_.front();
                                 pndVars_.pop_front();
                            }

                            if(specifiedName)
                            {
                               // Sometimes gdb produces a name="D10" kind of field in the ^done,value= lines,
                               // and sometimes it doesn't.
                               //
                               // When it does, lets verify that the name matches the variable we are 
                               // expecting process

                               string specNam = specifiedName->value();

                               if(!name.empty())
                               {
                                  if(specNam != name)
                                  {
                                     DIAGNOSTIC_LOG_BLOCK
                                     {
                                         CXXTLS_LOG("Error:  gdb display variable name, " 
                                                << specNam 
                                                << ", doesn't match expected value:  " 
                                                << name
                                               );
                                     }
                                  }
                               }
                               else
                                   name = specNam; // should never get here


                            }

                            DIAGNOSTIC_LOG_BLOCK
                            {
                                 CXXTLS_DEBUG("Updating display variable '" << name << "=" << value << "'");
                            }

                            VariableInfo* where = findOrCreateMonitoredVariable(name);

                            if(!where)
                            {
                               DIAGNOSTIC_LOG_BLOCK
                               {
                                    CXXTLS_DEBUG("Error:  unable to find / create variable: '" << name << "'");
                               }
                            }
                            else
                            {

                                VarObj* type_value  = c->findChild("type");
                                
                                bool listChildren=true;
                                
                                if(type_value)
                                {
                                   string typeText = type_value->value();
                                
                                   char lastChar = ' ';
                                
                                   if(!typeText.empty())
                                   {
                                       lastChar = typeText[typeText.size()-1];
                                   }
                                
                                
                                   if(    lastChar == ']' 
                                      ||  lastChar == '*'
                                     )
                                   {
                                       listChildren=false;   // don't recurse on the subtypes of arrays and pointers
                                   }
                                
                                   where->type_ = typeText;
                                
                                   static string charConstPointer("const char *");
                                   static string charPointer("char *");
                                
                                   if(    typeText == charConstPointer
                                      ||  typeText == charPointer
                                     )
                                   {
                                      // remove the leading address from the value that gdb prints
                                      // for string pointers
                                
                                      size_t valueOffset = value.find_first_not_of("0x123456789abcdefABCDEF ");
                                
                                      value = value.substr(valueOffset);
                                
                                   }
                                
                                }
                                
                                where->value_ = done_value->value();
                                
                                VarObj* numChild = c->findChild("numchild");
                                
                                if(listChildren && numChild)
                                {
                                   string numchildString = numChild->value();
                                
                                   if(numchildString != "0")
                                   {
                                      sendCommand( string("-var-list-children ") + name );
                                   }
                                }

                                where->incrementChangeCount();

                            }

                        }
                        else
                        {
                            ScopedMutexLock lock(stkMut_);

                            if(expWat_)
                            {
                                expWat_ = false;
                                expVal_ = done_value->value();
                                ++expCnt_;
                            }
                            else
                            {
                                DIAGNOSTIC_LOG_BLOCK
                                {
                                     CXXTLS_DEBUG("Error:  Why aren't there any pending variable names?");
                                }
                            }
                        }

                    } // ^done,value 

                    if(bkpt.get())
                    { // ^done,bkpt

                        CompositeVarObj *info = dynamic_cast<CompositeVarObj*>(bkpt.get());

                        if(info)
                        {
                            // we did in fact get an composite object

                            VarObj* number = info->findChild("number");

                            if(number)
                            {
                                string numberValue = number->value();

                                ScopedMutexLock lock(mutex_);

                                Breakpoints::iterator where = breakpoints_.find(numberValue);

                                if(where == breakpoints_.end())
                                {
                                    mrBreakpoint_ = numberValue;

                                    breakpoints_.insert( Breakpoints::value_type(numberValue, bkpt.release()) );
                                    DIAGNOSTIC_LOG_BLOCK
                                    {
                                         CXXTLS_DEBUG("Adding breakpoint " << numberValue);
                                    }
                                }
                                else
                                {
                                    delete where->second;
                                    where->second = bkpt.release();
                                }

                            }
                            else
                            {
                                DIAGNOSTIC_LOG_BLOCK
                                {
                                     CXXTLS_DEBUGNONL("Error:  no 'number' field in bkpt={...}:  " << line);
                                }
                            }
                        }
                        else
                        {
                            DIAGNOSTIC_LOG_BLOCK
                            {
                                 CXXTLS_DEBUGNONL("Error:  expected an bkpt={...} but found something else: " << line);
                            }
                        }

                    }  // ^done,bkpt=

                    if(bpTable.get())
                    { // ^done,BreakpointTable

                       // the BreakpointTable is a list of the breakpoints and there current state.
                       // It has a bunch of parameters, among them a hdr={}.  However we are interested
                       // in the body=[] which contains a set of bkpt={} objects in it.
                       //
                       // We are going to replace the breakpoint list of var objects with copies of the varobjs
                       // in the body list of bkpts

                        CompositeVarObj *info = dynamic_cast<CompositeVarObj*>(bpTable.get());

                        if(info)
                        {
                            ArrayVarObj *body = dynamic_cast<ArrayVarObj*>(info->findNestedChild("body"));

                            if(body)
                            {
                                size_t count = body->size();

                                for(size_t i = 0; i < count; ++i)
                                {
                                    CompositeVarObj *var = dynamic_cast<CompositeVarObj*>( (*body)[i] );

                                    if(var)
                                    {
                                        CompositeVarObj *bkpt = dynamic_cast<CompositeVarObj*>(var->findNestedChild("bkpt"));

                                        if(bkpt)
                                        {
                                            VarObj *number = bkpt->findNestedChild("number");

                                            if(number)
                                            {
                                                string numberValue = number->value();
                                            
                                                VarObj::ParseResult result = VarObj::parse(bkpt->value());

                                                if(result.rv_)
                                                {
                                                     ScopedMutexLock lock(mutex_);
                                                     
                                                     std::auto_ptr<VarObj> deleter(result.rv_);  // make sure memory is cleaned up
                                                     
                                                     Breakpoints::iterator where = breakpoints_.find(numberValue);
                                                     
                                                     if(where == breakpoints_.end())
                                                     {
                                                         breakpoints_.insert( Breakpoints::value_type(numberValue, deleter.release()) );
                                                     }
                                                     else
                                                     {
                                                         delete where->second;
                                                         where->second = deleter.release();
                                                     }
                                                }
                                                else
                                                {
                                                    DIAGNOSTIC_LOG_BLOCK
                                                    {
                                                         CXXTLS_DEBUGNONL("Error:  BreakpointTable.body[" << i << "].bkpt produces mangled value()");
                                                    }
                                                    break;
                                                }
                                            }
                                            else
                                            {
                                                 DIAGNOSTIC_LOG_BLOCK
                                                 {
                                                      CXXTLS_DEBUGNONL("Error:  BreakpointTable.body[" << i << "].bkpt.number is missing");
                                                 }
                                                 break;
                                            }

                                        }
                                        else
                                        {
                                            DIAGNOSTIC_LOG_BLOCK
                                            {
                                                 CXXTLS_DEBUGNONL("Error:  expected an BreakpointTable.body[" << i << "].bkpt to be a CompositeVarObj -- it is not");
                                            }
                                            break;
                                        }
                                    }
                                    else
                                    {
                                        DIAGNOSTIC_LOG_BLOCK
                                        {
                                             CXXTLS_DEBUGNONL("Error:  expected an BreakpointTable.body[" << i << "] to be a CompositeVarObj -- it is not");
                                        }
                                        break;
                                    }
                                }
                               
                            }
                            else
                            {
                                 DIAGNOSTIC_LOG_BLOCK
                                 {
                                      CXXTLS_DEBUGNONL("Error:  expected an BreakpointTable.body to be an ArrayVarObj -- it is not");
                                 }
                            }
                            
                        }
                        else
                        {
                            DIAGNOSTIC_LOG_BLOCK
                            {
                                 CXXTLS_DEBUGNONL("Error:  expected an BreakpointTable to be a CompositeVarObj -- it is not");
                            }
                        }
                       
                    } // ^done,BreakpointTable

                    if(children)
                    {
                        // "children" is an array of composite objects containing a single member called "child" which
                        // defines various features of child members of gdb variable object.
                        //
                        // Children have a "name" which is formatted like this:
                        //
                        //    "<varObjectName>.(public|private)[.memberExpression]"
                        //
                        // Where a member expression is defined like a child name, recursively.
                        //

                        ArrayVarObj &kids = *children;
                        size_t       cnt  = kids.size();

                        for(size_t i = 0; i < cnt; ++i)
                        {
                            CompositeVarObj *kid = dynamic_cast<CompositeVarObj*>(kids[i]);

                            if(!kid)
                            {
                                DIAGNOSTIC_LOG_BLOCK
                                {
                                    CXXTLS_LOG("Error:  invalid children= tag from gdb, member child is missing\n"
                                           "       " << line);
                                }
                                break;
                            }

                            CompositeVarObj *child = dynamic_cast<CompositeVarObj*>(kid->findChild("child"));

                            if(!child)
                            {
                                DIAGNOSTIC_LOG_BLOCK
                                {
                                    CXXTLS_LOG("Error:  invalid children= tag from gdb, member child is wrong varobj type\n"
                                           "       " << line);
                                }
                                break;
                            }

                            VarObj *childName = child->findChild("name");
                            VarObj *exp       = child->findChild("exp");
                            VarObj *numchild  = child->findChild("numchild");
                            VarObj *typePtr   = child->findChild("type");     // optional

                            string typeName;  

                            if(typePtr)
                            {
                                typeName = typePtr->value();
                            }

                            if(!childName || !exp || !numchild)
                            {
                                DIAGNOSTIC_LOG_BLOCK
                                {
                                     CXXTLS_LOG("Error:  invalid children= tag from gdb, member child is missing fields\n"
                                        << "       :  name\n" 
                                        << "       :  exp\n"
                                        << "       :  numchild\n"
                                        << "       " << line);
                                }

                                break;
                            }

                            DIAGNOSTIC_LOG_BLOCK
                            {
                                  CXXTLS_DEBUG("Found CHILD ATRIB:  name     " << childName->value());
                                  CXXTLS_DEBUG("                 :  exp      " << exp->value());
                                  CXXTLS_DEBUG("                 :  type     " << typeName);
                                  CXXTLS_DEBUG("                 :  numchild " << numchild->value());
                            }
                            updateMonitoredVariableAttributes(childName->value(), exp->value(), typeName);

                            if(numchild->value() != "0")
                            {
                                // if this type has nested sub-definitions, go get them
                                // except for the case of arrays -- we don't want to deal with
                                // the child elements like this.  After all, you could have an 
                                // array with millions of elements.

                                bool listSubChildren = true;

                                if(!typeName.empty())
                                {
                                   char lastChar = typeName[ typeName.size()-1 ];

                                   if(   lastChar == ']'
                                      || lastChar == '*'
                                     )
                                   {
                                      listSubChildren = false; // don't recurse on subtypes of arrays and ptrs
                                   }
                                }

                                if(listSubChildren)
                                    sendCommand( string("-var-list-children ") + childName->value());
                            }


                        }

                    } // children=

                    if(stack.get())
                    {
                       // handle a non-empty stack 

                       ScopedMutexLock lock(mutex_);

                       stack_.reset( stack.release() );

                    }

                    if(args.get())
                    {
                        // Copy out the args array from the frame in the level output by gdb.  The request
                        // for function arguments always is a request for exactly one stack frame, so we 
                        // will not worry about others.  See stackFrameParms()



                        ScopedMutexLock lock(stkMut_);

                        stkPrm_ = "";



                        ArrayVarObj *frameVec = dynamic_cast<ArrayVarObj*>(args.get());

                        if(frameVec)
                        {
                            VarObj *firstFrame = (*frameVec)[0];

                            VarObj *argInfo = firstFrame->findNestedChild("frame/args");
                            
                            ArrayVarObj* vec = dynamic_cast<ArrayVarObj*>(argInfo);
                            
                            if(vec)
                            {
                               stkPrm_ = vec->value();
                               ++stkCnt_;
                            }
                            
                        }
                    }

                    if(variables.get())
                    {
                        // Copy out the variables array


                        ScopedMutexLock lock(stkMut_);

                        stkPrm_ = "";

                        stkVar_ = variables->value();

                        ++varCnt_;
                    }



                } // cur line has valid parsable list on it  after ^done,....

            }   // found ^done in the gdb output line

            incDoneErrorCount(); // and set error to empty string

            delete rval.rv_;
        }
        else
        if(firstWord == "^error")
        {
            //////////////////////////////////////////
            // We have given gdb a bad command or bad       
            // data to a command
            //////////////////////////////////////////

            char const *f = tmp.data() + comma+1;
            char const *l = tmp.data() + tmp.size();

            VarObj::ParseResult  rval = VarObj::parseList(f,l);

            if(rval.rv_ == 0)
            {
                DIAGNOSTIC_LOG_BLOCK
                {
                     CXXTLS_DEBUG("Error:  cannot parse:  " << line << "Error: " << rval.error_);
                }

                signalDisplayExpressionError("Unknown error");

            }
            else
            {
                CompositeVarObj* c = dynamic_cast<CompositeVarObj*>(rval.rv_);

                if(!c)
                {
                     DIAGNOSTIC_LOG_BLOCK
                     {
                          CXXTLS_DEBUG("Error:  cannot parse:  " << line << "Error: " << rval.error_);
                     }

                    signalDisplayExpressionError("Unknown error");

                }
                else
                {
                    // Ok, at this point we have parsed a list of VarObjs from the ^error,... line.
                    //
                    // The variable "c" now refers to the CompositeVarObj that holds the parsed
                    // components.

                    std::auto_ptr<VarObj> msg    (c->release("msg"));

                    std::string msgValue;

                    if(msg.get())
                    {
                       msgValue = msg->value();

                       if(!msgValue.empty())
                       {
                          // ^error,msg="..."

                         if(!pndVars_.empty())
                         {
                              CXXTLS_LOG("Error:  gdb error: " << msg->value() << " for display var: " << pndVars_.front());
                              pndVars_.pop_front();
                         }
                         else
                         DIAGNOSTIC_LOG_BLOCK
                         {
                             CXXTLS_LOG("Error:  gdb error: " << msg->value());
                         }

                         signalDisplayExpressionError(msg->value());

                         // ^error,msg="The program is not being run."

                         if(msgValue == "The program is not being run.")
                         {
                            setState(terminated, "The program is not being run");
                         }

                       }

                       incDoneErrorCount(msgValue);
                    }
                }
            }
        }
        else
        if(firstWord == "=thread-group-started")
        {
            //////////////////////////////////////////
            // the target process has started running
            //////////////////////////////////////////

            char const *f = tmp.data() + comma+1;
            char const *l = tmp.data() + tmp.size();

            VarObj::ParseResult rv = VarObj::parseList(f,l);

            std::auto_ptr<VarObj>  info(rv.rv_);

            if(info.get())
            {
                VarObj *pidInfo = info->findNestedChild("pid");

                if(pidInfo)
                {
                   pid_ = lexical_cast<pid_t>(pidInfo->value());
                   setState(running, "thread-group-started");

                }

            }

            
        }
        else
        if(firstWord == "=thread-group-exited")
        {
            setState(terminated, "thread-group-exited");
           
        }
    }
    else
    if(   strlen(line) > 3 
       && line[0] == '~' 
       && line[1] == '"'
      )
    {
         bool scoopMode = false;

         {
            ScopedMutexLock lock(doneErrorMutex_);

            scoopMode = doneErrorScooping_;
         }

         // handle gdb otuput lines that look like this:
         //
         //  ~"...."
         //

         // NOTE NOTE NOTE NOTE NOTE!
         //
         //  If you are going to add lines to interpret
         //  here, recognize that the buffer, line, 
         //  contains an extra newline over and above what
         //  you think you see by looking at the gdb output
         //  log (gdblog.log or gdb output you see when you
         //  run it directly).
         //
         //  You must account for this trailing newline
         //  in any strings that you compare to line....
         //
         //  This has burned me every time that I added
         //  a new one...

         if(scoopMode)
         {
            // we are in a mode wherein all lines beginning with ~"" get scooped up
            // and appended to doneScooped_.

            std::string output = line;

            if(!output.empty() && output[ output.size()-1 ] == '\n')
            {
                output.erase( output.size()-1 );

                output.erase(0,2);

                if(!output.empty() && output[ output.size()-1 ] == '"')
                   output.erase(output.size()-1);

                if(output.size() > 1)
                {
                   if(output[0] == '\\' && output[1] == 'n')
                      output.erase(0,2);
                }

                if(output.size() > 2)
                {
                   size_t s = output.size();

                   if(output[s-1] == 'n' && output[s-2] == '\\')
                   {
                      output.erase(s-2,2);
                   }
                }


                ScopedMutexLock lock(doneErrorMutex_);

                doneScooped_.push_back(output);
            }


         }
         else
         if(0 == strcmp(line, "~\"Continuing.\\n\"\n"))
         {
            // ignore this line -- specifically, do not
            // call accumulateDisplayExpressionFragment on it
         }
         else
         if(0 == strcmp(line,"~\"\\n\"\n"))   //  ~"\n"
         {
            // ok, we have hit the encoded end-of-line
            // which means that a display variable, or some other
            // kind of display object has been completely printed.
            //
            // This code will concatenate the object into one long
            // line then will feed that line to a function that will
            // decide whether or not to use it or ignore it.

            recordDisplayExpressionResults();

         }
         else
         if(0 == strcmp(line, "~\"(no debugging symbols found)...done.\\n\"\n"))
         {
            ScopedMutexLock lock(mutex_);

            hasSymbols_ = false; // we are now known not to have symbols
         }
         else
         if(0 == strncmp(line, "~\"SIG", 5))
         {
             std::string tmp;

             char const *scan = line + 2;  // skip ~"

             while(*scan)
             {
                char c = *scan;

                if(c == '\\')
                {
                   ++scan;

                   c = *scan;

                   if(c == 0 || c == 'n')
                      break;  // end of string or end of line..

                   ++scan;

                   switch(c)
                   {
                       default:  tmp += c; break;

                       case 'r':  tmp += '\r'; break;
                       case 'n':  tmp += '\n'; break;
                       case 'e':  tmp += 0x1b; break;
                       case 't':  tmp += '\t'; break;
                       case 'a':  tmp += 0x07; break;
                       case 's':  tmp += ' ';  break;

                   }
                }
                else
                {
                  if(c == '"')
                    break;
                   tmp += c;
                   ++scan;
                }
             }

             setSignalHandling(tmp);
            
         }
         else
         if(0 == strncmp(line,"~\"Working directory ", 20))
         {

            std::string tmp = line;

            size_t directory = tmp.find("directory ");

            if(directory < tmp.size())
            {
                size_t dirStart = directory + 10;

                if(dirStart < tmp.size() - 2 )
                {
                   tmp = tmp.substr(dirStart);

                   ScopedMutexLock lock(mutex_);

                   size_t length = tmp.find(".\\n");

                   tmp = tmp.substr(0, length);

                    curDir_ = tmp;
                   ++curDirCnt_;

                }

            }


         }
         else
         if(line[0] == '~' && line[1] == '"')
         {
             GDBLOG("Found a ~\"...\" line =>>" << line);

             if(    (line[3] == 'y')
                &&  (line[4] == 'p')
                &&  (line[5] == 'e')
                &&  (line[6] == ' ')
               )
             {
                 // line looks like this: 
                 //
                 //    &"type = TYPENAME"\n
                 //
             
                 std::string rv = line;
             
                 rv.erase(0,9);
             
                 size_t lastQuote = rv.find_last_of('\\');  // remove trailing \n"\n
             
                 if(lastQuote < rv.size())
                 {
                     rv.erase(lastQuote, rv.size()-lastQuote);
                 }
             
                 ScopedMutexLock lock(doneErrorMutex_);
             
                 doneWhatisStr_ = rv;
             
                 ++doneErrorCnt_;
             
                 
             
             }
             else
                 accumulateDisplayExpressionFragment(line);  
                                           // record this fragment of a string object for
                                           // future use when an encoded end-of-line occurs
                                           // see the other half of this f statement
         }
         // put no more else clauses here!

    }   // handle ^"...

   
}
//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
recordDisplayExpressionResults()
{
   ScopedMutexLock lock(eMutex_);

   size_t size = eaccum_.size();

   size_t colon = eaccum_.find_first_not_of("0123456789");

   if(    colon < size
      &&  eaccum_[colon] == ':'
     )
   {
       // we have "number: ..."

       evalues_.push_back(eaccum_);
   }
   else
   {
      // the accumulator has trash in it

      GDBLOG("recordDisplayExpressionResults():  ignoring ==>" << eaccum_ << "<==");

   }

   eaccum_.resize(0);


   ++eChgCnt_;

}

//=============================================================================
//=============================================================================
//=============================================================================

void 
GdbController::
accumulateDisplayExpressionFragment(std::string const &inputValue) // value of 1 line of gdb output
{
   std::string value = inputValue;

   size_t size = value.size();

   size_t lastNonTerminator = value.find_last_not_of("\r\n \t");

   if(lastNonTerminator < size)
     {
       value.substr(0, lastNonTerminator+1);
       size = value.size();
     }

   size_t firstQuote = value.find_first_of('"');
   size_t lastQuote  = value.find_last_of('"');
   
   if(   firstQuote >= size
      || lastQuote  >= size
     )
   {
      GDBLOG("accumulateDisplayExpressionFragment():  ERROR:  Unbalanced double quotes");

      return;  // did not find the necessary ~"..."  as expected
   }

   value = value.substr(firstQuote+1, lastQuote-(firstQuote+1));

   {
      // unlike the command -expression-evaluate, there are some bugs in the automatic
      // display of "display varname" style output.  Specifically, sometimes there are
      // extraneous  \ sequences.  Remove them here

      std::string tmp;

      char const *scan = value.data();
      char const * end = scan + value.size();

      while(scan != end)
      {
          char c = *scan;

          if(c == '\\')
          {
              ++scan;

              if(scan == end)
              {
                  tmp += c;
                  break;
              }

              c = *scan++;

              switch(c)
              {
                  case 'a':
                  case 'r':
                  case 'e':
                  case 'b':
                  case 'n':
                  case 't':
                      tmp += '\\';
                      tmp += c;
                      break;


                  default: tmp += c;
                  break;
              }

          }
          else
          {
             tmp += c;
             ++scan;
          }
      }


      value = tmp;

   }


   ScopedMutexLock lock(eMutex_);

   if(!eaccum_.empty())
        eaccum_ += ' ';

   eaccum_ += value;

   GDBLOG("accumulateDisplayExpressionFragment():  current accumulator ==>" << eaccum_);

}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::sendArgsAndSetup()
{

   string command;



   if(firstSetArgs_)
   {

     bool superseedDir = curDir_.empty(); // save file superseeds the defalt

       if(curDir_.empty())
       {

           if(!cdForRun_.empty())
           {
              // before calling run(), the user has set cdForRun_

              curDir_ = cdForRun_;

              superseedDir = false; // user has superseed save file and default
           }
           else
           {
               // user hasn't superseeded the start directory for the
               // run, so lets start in the same directory as the
               // executable.
               
               FileName tmp(exe_);
               
               curDir_ = tmp.dirname();
               
               if(!curDir_.empty())
                  curDir_.erase(curDir_.size()-1); 
               
           }

       }

       firstSetArgs_ = false;

       command= string("file ")     + exe_  + "\n";
       sendCommand(command);

       bool readBreakpoints = false;

       if(!saveFile_.empty())
       {
               VarObj::ParseResult rv = VarObj::readAndParseFile(saveFile_);

               if(rv.rv_)
               {
                   ArrayVarObj *breakpoints = dynamic_cast<ArrayVarObj*>(rv.rv_->findChild("breakpoints"));

                   if(breakpoints)
                   {
                       size_t size = breakpoints->size();

                       for(size_t i = 0; i < size; ++i)
                       {
                           VarObj *bp = (*breakpoints)[i];

                           if(bp)
                           {
                               VarObj *fullname = bp->findChild("fullname");
                               VarObj *line     = bp->findChild("line");

                               if(fullname && line)
                               {
                                   if(!readBreakpoints)
                                   {
                                      pndBP_.clear();  // discard the single breakpoint set on main() that is automatically
                                                       // added by the system as a whole.  
                                                       //
                                                       // We do this to prevent a gradual build up of the number of saved breakpoints
                                                       // all pointing to main.
                                                       //
                                                       // Note that if there are no breakpoints saved int the config file from a previous session,
                                                       // we will never get here and thus will never clear the  breakpoint on main set
                                                       // elsewhere in the code.

                                      readBreakpoints = true;
                                   }
                                   addBreakpoint(fullname->value(), lexical_cast<size_t>(line->value()));
                               }
                           }
                       }

                   }  // handling breakpoints

                   if(args_.empty())
                     {

                           VarObj *args = rv.rv_->findChild("args");
                           
                           if(args)
                           {
                              args_ = args->value();
                           }
                     }

                   if(superseedDir)
                     {

                           VarObj *dir = rv.rv_->findChild("dir");
                           
                           if(dir)
                           {
                              curDir_ = dir->value();
                           }
                     }


               }  // found a parsable VarObj

           
       }  // reading save file from previous run


   }

   command= string("set args ") + args_ + "\n";

   sendCommand(command);

       CXXTLS_FOREACH(std::string const &e, pndBP_)
       {
           sendCommand( string("-break-insert ") + e );
       }

   pndBP_.clear();


   sendCommand("set print elements 0");

   sendCommand(std::string("cd ") + curDir_);

   dirty_ = false;
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
adoptStopInfo(VarObj*p)
{
    ScopedMutexLock lock(mutex_);

    stopInfo_.reset(p);
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
clearStopInfo()
{
    ScopedMutexLock lock(mutex_);

    stopInfo_.reset(0);
}

//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
stopInfo() const
{
    ScopedMutexLock lock(  const_cast<Mutex&>(mutex_)  );

    if(stopInfo_.get())
    {
        return stopInfo_.get()->value();
    }

    return "";
}

//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
stack() const
{
   ScopedMutexLock lock(  const_cast<Mutex&>(mutex_)  );

   if(stack_.get())
   {
       return stack_.get()->value();
   }

   else return "";
}

//=============================================================================
//=============================================================================
//=============================================================================
void 
GdbController::
threads(GdbController::IdList &ids) const
{
    ids.clear();

    ScopedMutexLock lock(  const_cast<Mutex&>(mutex_)  );

    ArrayVarObj const *threads = dynamic_cast<ArrayVarObj*>(threads_.get());

    if(threads)
    {
        size_t count = threads->size();

        for(size_t i = 0; i < count; ++i)
        {
            CompositeVarObj const *thread = dynamic_cast<const CompositeVarObj*>((*threads)[i]);

            if(thread)
            {
               VarObj const *id = thread->findChild("id");

               if(id)
               {
                   ids.push_back(id->value());
               }
            }

        }
    }
}
//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
thread(std::string const &threadId) const
{
   // find the thread whose id is specified and return a CompositeVarObj encoded
   // description of it.

   if(threads_.get())
   {
       ArrayVarObj const *threadList = dynamic_cast<ArrayVarObj const*>(threads_.get());

       size_t count = threadList->size();

       for(size_t i = 0; i < count; ++i)
       {
          CompositeVarObj const *thread = dynamic_cast<CompositeVarObj const*>((*threadList)[i]);

          if(thread)
          {
             VarObj const *id = thread->findChild("id");

             if(id)
             {
                 if(id->value() == threadId)
                 {
                     return thread->value();  
                 }
             }
          }

       }

   }

   return std::string();
}
//=============================================================================
//=============================================================================
//=============================================================================
VarObj const *
GdbController::
changelist() const
{
    ScopedMutexLock lock(  const_cast<Mutex&>(mutex_)  );

    return changelist_.get();
}

//=============================================================================
//=============================================================================
//=============================================================================
std::string const &
GdbController::
curThread() const
{
    ScopedMutexLock lock(  const_cast<Mutex&>(mutex_)  );

    return curThread_;
}


//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
changeExpression(std::string const &e, std::string const &value)
{
    pndVars_.push_back(e);
    sendCommand( string("-var-assign ") + e + " " + value);
}

//=============================================================================
//=============================================================================
//=============================================================================

static Mutex  m;
static int    nextUnique=10;  // start with 10 to get a >2 digit unique id

std::string 
GdbController::
nextId(char type)
{
    ScopedMutexLock lock(m);

    char buffer[246];

    snprintf(buffer, sizeof(buffer), "%c%d", type, nextUnique++);

    return buffer;

}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
variables(GdbController::IdList &names) const
{
    names.clear();

    CXXTLS_FOREACH(VariableValues::value_type const &cur, variables_)
    {
       if(cur.second.inScope_)
           names.push_back(cur.first);
    }
}

//=============================================================================
//=============================================================================
//=============================================================================
GdbController::VariableInfo const *
GdbController::
variable(std::string const &id) const
{
     VariableValues::const_iterator where = variables_.find(id);

     if(where != variables_.end())
     {
        return &where->second;
     }


     return 0;
}


//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
breakpoints(std::list<std::string> &ids) const
{
    ScopedMutexLock lock(  const_cast<Mutex&>(mutex_) );

    ids.clear();

    CXXTLS_FOREACH(Breakpoints::value_type const &cur, breakpoints_)
    {
        ids.push_back(cur.first);
    }
}

//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
breakpoint(std::string const &id) const
{
    Breakpoints::const_iterator where = breakpoints_.find(id);

    if(where != breakpoints_.end())
    {
       return where->second->parsable_value();
    }

    return std::string();
}



//=============================================================================
//=============================================================================
//=============================================================================
std::string 
GdbController::
removeBreakpoint(std::string const &id)
{
     ScopedMutexLock lock(mutex_);

     Breakpoints::iterator where = breakpoints_.find(id);

     if(where != breakpoints_.end())
     {
         breakpoints_.erase(where);

         lock.unlock();   // release the mutex_ so we don't get hung somewhere
                          // in the controlReader process.  

         size_t startCount = clearDoneErrorMessage();

         sendCommand( string("-break-delete ") + id);

         return awaitDoneErrorChange(startCount);

     }

     return "invalid breakpoint id";
}

//=============================================================================
//=============================================================================
//=============================================================================
void 
GdbController::
updateMonitoredVariableAttributes(std::string const &name,
                                  std::string const &exp,
                                  std::string const &type
                                 )
{
    if(name.empty())
       return;


    StrTool::stringlist_t nameComponents;

    StrTool::Is_Delim dot('.');

    int rc = StrTool::parse_words(name, &nameComponents, 20, dot);

    if(rc == 0)
       nameComponents.push_back(name); 

    string varName = nameComponents.front();

    nameComponents.pop_front();

    VariableValues::iterator var = variables_.find(varName);

    if(var == variables_.end())
    {
        VariableInfo newInfo;

        variables_[varName] = newInfo;

        var = variables_.find(varName);
    }

    VariableInfo *scan = &var->second;

    // at this time, scan points to a variable info object and nameComponents
    // points to information describing a child thereof.  We will search down
    // from the top of the variable tree, as pointed to by scan for members
    // whose names appear in nameComponents.front().  Whenever we run out of
    // name components, we stop and assign expr and type to the variable of
    // interest.

    string parent = varName;

    while(!nameComponents.empty())
    {
       string curComponent = nameComponents.front();

       nameComponents.pop_front();

       string childId = parent + "." + curComponent;

       VariableValues::iterator where = scan->children_.find(curComponent);

       if(where == scan->children_.end())
       {
           scan->children_.insert( VariableValues::value_type(curComponent, VariableInfo()) );

           where = scan->children_.find(curComponent);

           where->second.id_ = childId;
       }

       if(nameComponents.empty())
       {
          // we are currently pointing to the last item

          where->second.expr_ = exp;
          where->second.type_ = type;

          return;

       }

       parent = childId;
       scan   = &where->second;




    }





}
//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
logMonitoredVariables() const
{
    DIAGNOSTIC_LOG_BLOCK
    {
          CXXTLS_LOG("GdbController::logDisplayExpressions:");

          string linePrefix="  ";
          
          CXXTLS_FOREACH(VariableValues::value_type const &cur, variables_)
          {
              CXXTLS_LOG(cur.first  << ":");
              cur.second.log(linePrefix);
          }
    }
    
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::VariableInfo::
log(std::string const &linePrefix) const
{

    {  
        string nesting = linePrefix;
        CXXTLS_LOG(   linePrefix << "Value\t:\t" << value_ << '\n'
               << linePrefix << "Type\t:\t"  << type_  << '\n'
               << linePrefix << "Expr\t:\t"  << expr_  << '\n'
               << linePrefix << "GdbId\t:\t" << id_  
              );


        nesting = linePrefix + "        ";
        
        CXXTLS_LOG("");
        
        CXXTLS_FOREACH(VariableValues::value_type const &cur, children_)
        {
            CXXTLS_LOG(linePrefix << cur.first  << ":");
            cur.second.log(nesting);
        }

    }

}
//=============================================================================
//=============================================================================
//=============================================================================
GdbController::VariableInfo *
GdbController::
findOrCreateMonitoredVariable(std::string const &gdId)
{

    if(gdId.empty())
    {
       return 0;
    }

    StrTool::stringlist_t nameComponents;

    StrTool::Is_Delim dot('.');

    int rc = StrTool::parse_words(gdId, &nameComponents, 20, dot);

    if(rc == 0)
       nameComponents.push_back(gdId); 

    string varName = nameComponents.front();

    nameComponents.pop_front();

    VariableValues::iterator var = variables_.find(varName);

    if(var == variables_.end())
    {
        VariableInfo newInfo;

        variables_[varName] = newInfo;

        var = variables_.find(varName);
    }

    VariableInfo *scan = &var->second;

    // at this time, scan points to a variable info object and nameComponents
    // points to information describing a child thereof.  We will search down
    // from the top of the variable tree, as pointed to by scan for members
    // whose names appear in nameComponents.front().  Whenever we run out of
    // name components, we stop and assign expr and type to the variable of
    // interest.

    string parent = varName;

    while(!nameComponents.empty())
    {
       string curComponent = nameComponents.front();

       nameComponents.pop_front();

       string childId = parent + "." + curComponent;

       VariableValues::iterator where = scan->children_.find(curComponent);

       if(where == scan->children_.end())
       {
           scan->children_.insert( VariableValues::value_type(curComponent, VariableInfo()) );

           where = scan->children_.find(curComponent);

           where->second.id_ = childId;
       }

       if(nameComponents.empty())
       {
          // we are currently pointing to the last item

         scan = &where->second;

          break;

       }

       parent = childId;
       scan   = &where->second;

    }

    return scan;
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
triggerMonitoredVariableEvaluation(std::string const &e)
{
    ScopedMutexLock var(mutex_);

    pndVars_.push_back(e);
    sendCommand( string("-var-evaluate-expression ") + e);
}

//=============================================================================
//=============================================================================
//=============================================================================
bool GdbController::
writeTargetStdin(std::string const &s)
{
    ChildState st = state();

    DIAGNOSTIC_LOG_BLOCK
    {
         CXXTLS_DEBUGNONL("GdbController::writeTargetStdin:  state=" << stateName() << ", " << s);
    }


    if(    st == stopped
        || st == running
      )
    {
        bool rc = (s.size() == size_t( write(pty_->masterFD(), s.data(), s.size()) ));

        if(!rc)
        {
           int t = errno;

           DIAGNOSTIC_LOG_BLOCK
           {
                CXXTLS_LOG("Error in writeTargetStdin() = " << strerror(t));
           }
        }
    }

    return false;
}

//=============================================================================
//=============================================================================
//=============================================================================

std::string
GdbController::
stackFrameParms(std::string const &frameNumber)
{
    static std::string space(" ");
    static std::string request("-stack-list-arguments 1 "); // verbosity level=1 which gives {name="name",value="value"} instead of
                                                            // level=0 which gives {"name","value"}
                                                            // or level=2 which gives full structure definitions inside value

    size_t startChangeCount = argsChangeCount();

    sendCommand(request + frameNumber + space + frameNumber);

    bool timeout =  awaitStackParmChange(startChangeCount);  // wait up to 5 seconds for the stack parm contents to change

    if(!timeout)
    {
       ScopedMutexLock lock(stkMut_);
       return stkPrm_;  // varobj encoded list of name/value pairs definining function parameters
    }

    return std::string();
}

//=============================================================================
//=============================================================================
//=============================================================================

std::string
GdbController::
stackFrameVariables(std::string const &frameNumber)
{
    bool stackErr =  selectStackFrame(frameNumber);

    if(stackErr)
       return std::string();   //empty return means error


    static std::string space(" ");

    static std::string request("-stack-list-variables 1 "); // verbosity level=1 which gives {name="name",value="value"} instead of
                                                            // level=0 which gives {"name","value"}
                                                            // or level=2 which gives full structure definitions inside value

    size_t startChangeCount = variablesChangeCount();

    sendCommand(request);

    bool timeout =  awaitVariableChange(startChangeCount);  // wait up to 5 seconds for the stack parm contents to change

    if(!timeout)
    {
       ScopedMutexLock lock(stkMut_);
       return stkVar_;  // varobj encoded list of name/value pairs definining function parameters
    }

    return std::string();
}

//=============================================================================
//=============================================================================
//=============================================================================

std::string 
GdbController::
evaluateTmpExpression(std::string const &frame, std::string const &expInp)
{
    std::string exp = expInp;

    std::string exprType = whatis(frame,exp);

    if(!exprType.empty())
    {
       if(exprType[ exprType.size() -1 ] == '&')
       {
          exp = std::string("*&") + exp;
       }
    }


    bool stackErr = selectStackFrame(frame);

    if(stackErr)
    {
       return "";
    }


    std::string rv("-data-evaluate-expression ");

    rv += exp;

    size_t startCount = expressionChangeCount();

    {
       ScopedMutexLock lock(stkMut_);
       expWat_ = true;
    }

    sendCommand(rv);

    bool timeout = awaitExpressionChange(startCount);

    if(timeout)
        return "";

    return expressionValue(); // encoded string

}



//=============================================================================
//=============================================================================
//=============================================================================
void 
GdbController::
fetchDisplayExpressionResults(std::vector<std::string> &results)
{

    ScopedMutexLock lock(eMutex_);

    GDBLOG("Fetching display expression values for display:  count = " << evalues_.size());

    CXXTLS_FOREACH(std::string const &cur, evalues_)
    {
        GDBLOG("==> " << cur);
    }


    results = evalues_;


}

//=============================================================================
//=============================================================================
//=============================================================================
void 
GdbController::
fetchDisplayExpressions(std::map<std::string,  std::string> &expressions)
{
    ScopedMutexLock lock(eMutex_);

   GDBLOG("Fetching display expression names for display:  count = " << enames_.size());

    expressions = enames_;
    
}

//=============================================================================
//=============================================================================
//=============================================================================
bool 
GdbController::
displayExpressionResults()
{
    ScopedMutexLock lock(eMutex_);

    return !evalues_.empty();
    
}

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
createDisplayExpression(std::string const &frame, std::string const &exprInput)
{

    std::string exprType = whatis(frame, exprInput);   // whatis sets the stack frame for us

    std::string expression = exprInput;

    if(!exprType.empty())
    {
        if(exprType[ exprType.size() -1 ] == '&')
        {
           expression = std::string("*&") + expression;
        }
    }


    size_t cnt = displayVariableChangeCount();

    sendCommand(std::string("display ") + expression);

    return awaitDisplayVariableChange(cnt);  // return true if timeout


    
}

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
deleteDisplayExpression(std::string const &expression)
{

    sendCommand(std::string("undisplay ") + expression);

    sleep(1);

    return false;  // no timeout 


    
}


//=============================================================================
//=============================================================================
//=============================================================================

bool
GdbController::
awaitDisplayVariableChange(size_t startChangeCount, int seconds)
{
   int maxSeconds = seconds * 10;  // 10 shortNaps per second

   {
      ScopedMutexLock lock(eMutex_);

      eError_ = false;
   }

   for(int i = 0; i < maxSeconds; ++i)  
   {
      if(startChangeCount == displayVariableChangeCount())
         shortNap();  // sleep 0.1 seconds
      else
      {
         ScopedMutexLock lock(eMutex_);

         if(eError_)
         {
            GDBLOG("awaitDisplayVariableChange():  Error occured");
            return true;
         }
         return false;
      }
   }

   GDBLOG("Time out gdb to produce stack frame variables");

   return true; // if we get here, a time out has occurred
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
getSignalHandlers(SignalHandlers &output)
{
   ScopedMutexLock lock(mutex_);

   output = signalHandlers_;
}

//=============================================================================
//=============================================================================
//=============================================================================
void
GdbController::
setSignalHandling(std::string const &optionsString)
{
    SignalHandling tmp;

    StrTool::stringlist_t  words;

    StrTool::parse_words(optionsString, &words, 10000, StrTool::Is_Delim('\t'));

    if(words.size() >= 4)
    {
       // words.size() should be 5 but really, we have 'SIGNAME    YES' in the first string.

        StrTool::stringlist_t::iterator f = words.begin();

        size_t nameEnd = f->find_first_of(" \t");

        std::string name;

        name = f->substr(0, nameEnd);

        size_t yesno = f->find_first_not_of(' ', nameEnd);

        if(yesno > f->size())
           return;

        if( toupper((*f)[yesno]) == 'Y')
           tmp.stop_ = true;

        ++f;

        if(!f->empty() && toupper((*f)[0]) == 'Y')
           tmp.print_ = true;

        ++f;
        
        if(!f->empty() && toupper((*f)[0]) == 'Y')
           tmp.pass_ = true;

        ++f;
        
        tmp.description_ = *f;

        ScopedMutexLock lock(mutex_);

        SignalHandlers::iterator where = signalHandlers_.find(name);

        if(where == signalHandlers_.end())
           signalHandlers_[name] = tmp;
        else
           where->second = tmp;


    }

}

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
requestSignalHandlerList()
{
   {
       ScopedMutexLock lock(mutex_);

       signalHandlers_.clear();

   }

   sendCommand("info handle");

   sleep(1);

   if(signalHandlers_.empty())
      return true;

   return false;

}

// ~"Working directory /home/lboggs.\n"

//=============================================================================
//=============================================================================
//=============================================================================
bool 
GdbController::
awaitCwdChange(size_t startChangeCount, int seconds)
{
   int maxSeconds = seconds * 10;  // 10 shortNaps per second

   for(int i = 0; i < maxSeconds; ++i)  
   {
      if(startChangeCount == getCwdChangeCnt())
         shortNap();  // sleep 0.1 seconds
      else
         return false;
   }


   GDBLOG("Time out waiting for cwd state change");

   return true; // if we get here, a time out has occurred



}
//=============================================================================
//=============================================================================
//=============================================================================
std::string
GdbController::
getcwd()
{

    if(   state() != stopped
       && state() != terminated
      )
    {

        std::string sf = saveFile();

        if(!sf.empty())
        {
            VarObj::ParseResult info = VarObj::readAndParseFile(sf);

            if(info.rv_)
            {
                VarObj *dir = info.rv_->findChild("dir");

                if(dir)
                   return dir->value();
            }
        }

        if(exe_.empty())
        {
            return FileName::getcwd();
        }


        size_t lastSlash = exe_.find_last_of('/');

        if(lastSlash < exe_.size())
           return exe_.substr(0, lastSlash);

        return exe_;
         

    }

    size_t startCwdChangeCount = getCwdChangeCnt();

    sendCommand("pwd");

    if(awaitCwdChange(startCwdChangeCount))
    {
        return "";
    }

    return curDir_;

}

//=============================================================================
//=============================================================================
//=============================================================================
bool
GdbController::
setcwd(std::string const &newdir)
{

    if(   state() != stopped
       && state() != terminated
      )
    {
        cdForRun_ = newdir;
        curDir_   = newdir;
        return false; // no error
    }

    size_t startCwdChangeCount = getCwdChangeCnt();

    sendCommand(std::string("cd ") + newdir);

    return awaitCwdChange(startCwdChangeCount);

}

static bool isNo(std::string const &s)
{
   if(s.empty())
      return true;

   return s[0] == 'N' || s[0] == 'n';
}

//=============================================================================
//=============================================================================
//=============================================================================
bool 
GdbController::
handleSignal(std::string const &name,
                      std::string const &stop,
                      std::string const &print,
                      std::string const &pass
                     )
{
   std::string cmd = "handle ";

   cmd += name;  cmd += ' ';

   cmd += isNo(stop) ? "nostop" : "stop";

   cmd +=  ' ';

   cmd += isNo(print) ? "noprint" : "print";

   cmd +=  ' ';

   cmd += isNo(pass) ? "nopass" : "pass";

   cmd +=  ' ';


   sendCommand(cmd);

   shortNap();  

   return false;

}

//=============================================================================
//=============================================================================
//=============================================================================
std::string 
GdbController::
awaitDoneErrorChange(size_t startCount, int seconds)
{
   int maxIterations = 10 * seconds;   

   for(int i = 0; i < maxIterations; ++i)
   {
        if(doneErrorCount() != startCount)
        {
            ScopedMutexLock lock(doneErrorMutex_);

            std::string rv = doneErrorStr_;

            doneErrorStr_.resize(0);

            return rv;

        }

        shortNap(); // sleep 1/10 of a second
   }

   {
       ScopedMutexLock lock(doneErrorMutex_);

       if(doneErrorCnt_ != startCount)
       {
           std::string rv = doneErrorStr_;

           doneErrorStr_.resize(0);

           return rv;
       }
      
   }

   return "time out";

}

//=============================================================================
//=============================================================================
//=============================================================================

std::string 
GdbController::
whatis(std::string const &frame, std::string const &expr)
{
    bool stackError = selectStackFrame(frame);

    if(stackError)
        return "";  // emtpy string means error.

    size_t startCount;

    startCount = clearDoneErrorMessage(); // clears whatis string too

    sendCommand(std::string("whatis ") + expr);

    std::string error = awaitDoneErrorChange(startCount);

    if(!error.empty())
        return ""; // don't convey the error details but report a problem

    // we got no error

    ScopedMutexLock lock(doneErrorMutex_);

    return doneWhatisStr_;

}

//=============================================================================
//=============================================================================
//=============================================================================

bool 
GdbController::
doneErrorScoop(std::string const &command, std::vector<std::string> &output)
{
   output.clear();

   ChildState curState = state();

   if(   curState != stopped 
      && curState != terminated
     )
   {
      output.push_back(std::string("Error:  the target must be stopped or terminated to issue command:  ") + command);
      return true; // error
   }


   size_t startCount = 0;

   {
      ScopedMutexLock lock(doneErrorMutex_);
      doneErrorScooping_ = true;
      startCount = doneErrorCnt_;
      doneScooped_.clear();
   }

   sendCommand(command);

   std::string error = awaitDoneErrorChange(startCount,15);

   if(!error.empty())
   {
      output.push_back(error);

       ScopedMutexLock lock(doneErrorMutex_);
       doneErrorScooping_ = false;

      return true;  
   }

   {
       ScopedMutexLock lock(doneErrorMutex_);
       doneErrorScooping_ = false;
       doneScooped_.swap(output);
   }

   return false;

}

void GdbController::writeGdbLogMessage(std::string const &s)
   {
       GDBLOG(s);
   }


void GdbController::signalDisplayExpressionError(std::string const &msg)
{
    writeGdbLogMessage(msg);
    ScopedMutexLock lock(eMutex_);
    eError_ = true;
    ++eChgCnt_;
}

//=============================================================================
//=============================================================================
//=============================================================================


bool GdbController::selectStackFrame(std::string frame)
{
    size_t startCount = clearDoneErrorMessage(); // clears whatis string too

    sendCommand(std::string("-stack-select-frame ") + frame);

    std::string error = awaitDoneErrorChange(startCount); // -stack-select-frame triggers a ^done record in the log

    if(!error.empty())
       return true; // fail

    return false; // success

}

static Mutex readerThreadGuard;
static int   readerThreadCount_ =0;

void 
GdbController::
incrementReaderThreadCount()
{
    ScopedMutexLock lock(readerThreadGuard);
    ++ readerThreadCount_;
}
                                 
                                 

void 
GdbController::
decrementReaderThreadCount()
{
    ScopedMutexLock lock(readerThreadGuard);
    -- readerThreadCount_;
}
                            
                            

int 
GdbController::
readerThreadCount()
{
    ScopedMutexLock lock(readerThreadGuard);
    return readerThreadCount_;
}
                   
                   


//=============================================================================
//=============================================================================
//=============================================================================
} // namespace cxxtls

#else
   int please_ignore_gdbcontroller_cxx_on_windows=1;
#endif
