//
// Copyright 2012, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.


#include <cxxtls/patternmatch.h>
#include <string.h>
#include <memory>
#include <cxxtls/charset.h>
#include <iostream>
#include <stdlib.h>
#include <algorithm>



namespace cxxtls
{
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    typedef PatternMatcher::MatchType MatchType;

    struct PatternMatcher::Matcher
        ///  A Matcher is a component in a pattern match expression.  A full pattern
        ///  is a sequence of matchers.  The matchers are executed in sequence, only
        ///  if all of the matchers return true is the string considered a match.
        ///
    {
       virtual ~Matcher() {}

       static bool searchBackward(PatternMatcher::MatchType m)
       {
           return  (   m == PatternMatcher::SHORTEST_TRAILING 
                    || m == PatternMatcher::LONGEST_TRAILING
                   );
       }


       typedef PatternMatcher::const_iterator const_iterator;

       virtual bool findFirstMatch(const_iterator &f, const_iterator &l) const = 0;
          ///<  Given a character range, [f,l), search for the first match of this
          ///   pattern sub-expression.  Return true if a match is found that is 
          ///   completely within the region, [f,l), and adjust f and l to back the
          ///   match.  If false, is returned, do not modify f or l.

       virtual bool skipMatchingRegion(MatchType       t, 
                                       const_iterator &f, 
                                       const_iterator &l,
                                       bool            first,  // first node in the pattern
                                       bool            last    // last node in the pattern
                                      ) const =0;
           ///<  compare text in the range [f,l) to the pattern in this matcher.  If it matches,
           ///   move the f and possibly the l iterators and return true.  l is moved if we
           ///   are doing leading matches and l is moved if we are doing trailing matches.
           ///   If the function returns false, neither is moved.
           ///   
           ///   Based on the requested match type, behave differently.
           ///
           ///   When the match type is exact, the iterator f will only be moved if the
           ///   function returns true to indicate that a match occurred.
           ///
           ///   The final flags, first and last, are are used differently based on t.
           ///   They basically tell the function how to handle.  The flags should only
           ///   be true if the current invocation is on the first or last node in the
           ///   match expression, respectively.  The words first and last, in this
           ///   case refer to the left to right order in which the expressions appear in
           ///   the defining string.  

       virtual Charset finalCharSet() const = 0;
           ///<  Return a characterset containing the final character or characterset
           ///   specific to this matcher.   The 'final' character or charset refers
           ///   to the last character or character set in any sequence of characters
           ///   or character set.  Here are some examples:
           ///   
           ///       "abc"
           ///          ^                 the final charset contains only 'c'
           ///
           ///       "[a-z][m-n][xy]"
           ///                  ^^^^      the final charset contains 'x' and 'y'
           ///
           ///       "?"                
           ///        ^                   the final charset contains all valid characters
           ///    
           ///       "*"
           ///        ^                   the final charset contains all valid characters
           ///
           ///       "" (the final characterset is empty)

    };

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    struct FixedStringMatch  
        /// A literal text string matcher.  For example:
        ///
        ///    "abc.txt"
        ///
    :  public PatternMatcher::Matcher
    {
       std::string pattern_;     // fixed string which must be exactly matched

       FixedStringMatch(std::string const &m)
       : pattern_(m)
       {
       }

       virtual Charset finalCharSet() const
       {
           Charset rv;

           if(!pattern_.empty())
           {
              rv.set( pattern_[pattern_.size()-1], true );
           }

           return rv;
       }

       virtual bool findFirstMatch(const_iterator &f, const_iterator &l) const
       {
           const_iterator found =  std::search(f,l, pattern_.begin(), pattern_.end());

           if(found != l)
           {
                f = found;
                l = found + pattern_.size();
                return true;
           }

           return false;

       }


       virtual bool skipMatchingRegion(MatchType       t, 
                                       const_iterator &f, 
                                       const_iterator &l,
                                       bool            first,
                                       bool            last
                                      ) const
       {

           const_iterator m = pattern_.begin();
           const_iterator e = pattern_.end();

           const_iterator scan = f;

           if(!searchBackward(t))
           {
               // normal forwards search

               scan = f;

               while(scan != l && m != e)
               {
                  char cin = *scan;
                  char cm  = *m;
               
                  if(cin != cm)
                     return false;
               
                  ++scan;
                  ++m;
               }
               
               if(
                  (m != e)
                 )
                  return false;
               
               f = scan;
           }
           else
           {
              // reverse search

               scan = l;

               while(scan != f && e != m)
               {
                  --scan;
                  --e;

                  char cin = *scan;
                  char cm  = *e;
               
                  if(cin != cm)
                     return false;
               
               }
               
               if(
                  (e != m)
                 )
                  return false;

               l = scan;

           }


           return true;
       }

    };

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    struct AnyExistingCharMatch
        /// Match a squence of any characters of a fixed width.  For example:
        ///
        ///    "??"
        ///
    :  public PatternMatcher::Matcher
    {
       size_t freeCharacters_;  // number of characters in which any existing comparand character matches.

       AnyExistingCharMatch(size_t count)
       : freeCharacters_(count)
       {
       }

       virtual Charset finalCharSet() const
       {
           Charset rv;

           if(freeCharacters_)
           {
              rv.setAll();
           }

           return rv;
       }

       virtual bool findFirstMatch(const_iterator &f, const_iterator &l) const
       {
           // this kind of expression matches any characters but has a minimum field width.

           size_t size = l-f;

           if(size >= freeCharacters_)
           {
               l = f + freeCharacters_;

               return true;
           }

           return false;

       }

       virtual bool skipMatchingRegion(MatchType       t, 
                                       const_iterator &f, 
                                       const_iterator &l,
                                       bool            first,
                                       bool            last
                                      ) const
       {
           
           
           const_iterator scan;

           size_t remaining = freeCharacters_;

           if(!searchBackward(t))
           {
               scan = f;
               
               while(   (scan != l)
                     && (remaining > 0)
                    )
               {
                  ++scan;
                  --remaining;
               }
               
               if(
                  (remaining != 0)
                 )
               {
                  return false;
               }
               f = scan;
           }
           else
           {
               scan = l;
               
               while(   (scan != f)
                     && (remaining > 0)
                    )
               {
                  --scan;
                  --remaining;
               }
               
               if(
                  (remaining != 0)
                 )
               {
                  return false;
               }

               l = scan;
           }

           return true;
       }

    };

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    struct ZeroOrMoreMatch   
        /// Check for zero more characters of any value -- possibly ending with a specified
        /// terminating character.  For example:
        ///
        ///   "*"
        ///   "*.c"
        ///   "*x"
        ///
        /// Note that ZeroOrOneMatch has a prefix and suffix characterset, called "termset"
        /// and "priorset".  These are used for searching for the end of the match region.
        ///
    :  public PatternMatcher::Matcher
    {

       bool           searchForTermChar_;   // if true, then we need to stop at a termset
                                            // member

       bool           searchForPriorChar_;  // if true, then when searching backwards, we need
                                            // to stop at a member of the priorSet

       Charset        termset;   // a set of characters that terminate the search
       Charset        priorset;  // a set of characters that preceed the search region


       ZeroOrMoreMatch()
       :  searchForTermChar_(false)
       ,  searchForPriorChar_(false)
       {
       }

       ZeroOrMoreMatch(char termchar)
       :  searchForTermChar_(true)
       ,  searchForPriorChar_(false)
       {
          termset.set(termchar, true);
       }

       ZeroOrMoreMatch(Charset const &set)
       :  searchForTermChar_(true)
       ,  searchForPriorChar_(false)
       ,  termset(set)
       {
       }

       ZeroOrMoreMatch(Charset const &term, Charset const &prior)
       :  searchForTermChar_(true)
       ,  searchForPriorChar_(true)
       ,  termset(term)
       ,  priorset(prior)
       {
       }


       ZeroOrMoreMatch(bool priorSet, Charset const &set)
       :  searchForTermChar_(!priorSet)
       ,  searchForPriorChar_(priorSet)
       {
          if(priorSet)
              priorset = set;
          else
              termset  = set;
              
       }


       virtual Charset finalCharSet() const
       {
           Charset rv;

           rv.setAll();

           return rv;
       }

       virtual bool findFirstMatch(const_iterator &f, const_iterator &l) const
       {
           // this kind of expression matches any characters and has a variable
           // width field length.  It may or may not have a terminating character
           // set.
           //
           // If it does, then we search for the first character that appears
           // in that set.  If none do, then the entire range, [f,l), is
           // considered to be in the match.  If a match of the terminating
           // character is found, then it is not included in the field of
           // interest, but everything up to it is.
           //
           // If there is no terminating character, then the whole range matches.

           if(
                 (!searchForTermChar_)
              && (f != l)
             )
           {
               return true; // without a term char, it's all a match
           }

           if(searchForTermChar_)
           {
               const_iterator scan = f;

               while(scan != l)
               {
                   if(termset.test(*scan))
                   {
                      // we found the terminating character

                      l = scan;

                      return true;   // everything from [f,scan)

                   }

                   ++scan;
               }

           }

           return false;

       }

       virtual bool skipMatchingRegion(MatchType       t, 
                                       const_iterator &f, 
                                       const_iterator &l,
                                       bool            first,
                                       bool            last
                                      ) const
       {

           bool foundIt = false;
           const_iterator scan;

           if(!searchBackward(t))
           {

               if(    t == PatternMatcher::SHORTEST_LEADING
                  &&  !searchForTermChar_ 
                 )
               {
                   return true;    // "*" when used in the shortest leading request should return true but
                                   // not move the because that's how the bash string expression ${var#*} works.  This
                                   // bash expression outputs var without removing the entire contents.  
                                   // but if you use the longest, it will.
                                   //

               }

               scan = f;
               
               while(scan != l)
               {
                   unsigned int cur = *scan;
               
                   foundIt = termset.test(cur);
               
                   if(foundIt)
                   {
                       if(t == PatternMatcher::LONGEST_LEADING)
                       {
                           // search for the last match in the range

                           const_iterator later = scan;

                           ++later;

                          if(later != l)
                          {
                             
                              const_iterator lastFound = l;
                             
                              while(later != l)
                              {
                                  if(termset.test(*later))
                                  {
                                     lastFound = later;
                                  }
                             
                                  ++later;
                              }
                             
                              if(lastFound != l)
                              {
                                scan = lastFound;  // we found at least one other, later match
                                                   // and we are going to use it.
                              }
                          }
                       }

                       break;
                   }
               
                   ++scan;
               }
               
               if(   !foundIt 
                  && searchForTermChar_
                 )
               {
                  // we were looking for a terminating character
                  // and we did not find it
               
                  return false;
               }

               f = scan;
           }
           else
           {

               if(    t == PatternMatcher::SHORTEST_TRAILING
                  &&  !searchForPriorChar_ 
                 )
               {
                   return true;    // "*" when used in the shortest trailing request should return true but
                                   // not move the because that's how the bash string expression ${var%*} works.  This
                                   // bash expression outputs var without removing the entire contents.  
                                   // but if you use the longest.
                                   //
                                   // Note:  I considered return false if f == l on entry.  You can't match
                                   // nothing, I suppose.  However, this function is called as part of a larger
                                   // sequence and if it return false if f == l, then this pattern won't work
                                   // correctly:
                                   //
                                   //     "*x"
                                   //
                                   // If the comparand is "*", then return false here will cause a false mismatch
               }


               scan = l;
               
               while(scan != f)
               {
                   --scan;

                   unsigned int cur = *scan;
               
                   foundIt = priorset.test(cur);
               
                   if(foundIt)
                   {
                       if(t == PatternMatcher::LONGEST_TRAILING)
                       {
                          // see if there are any earlier matches for the prior set

                          const_iterator earlier = scan;

                          const_iterator lastFound = scan;

                          while(earlier != f)
                          {
                              --earlier;

                              if(priorset.test(*earlier))
                                 lastFound = earlier;
                          }

                          scan = lastFound; // may or may not have changed.


                       }

                       ++scan;  // don't move the final iterator back to the prior
                                // character.  For example:
                                //
                                //   b*  matches ab123
                                //                 ^^^  the '*' part,
                                //  but we need l to be pointing to 1 so that
                                //  when the 'b' part of the pattern expression
                                //  runs, it can know that the width is 1.
                       break;
                   }
               
               }
               
               if(   !foundIt 
                  && searchForPriorChar_
                 )
               {
                  // we were looking for a terminating character
                  // and we did not find it
               
                  return false;
               }

               l = scan;
           }
           return true;
       }

    };


    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    struct FixedCharsetMatch  
        ///  Match a sequence of characters, each of which must be a member of
        ///  a user defined character set.  See
        ///      "[8-z][m-z]"
        ///
    :  public PatternMatcher::Matcher
    {
       typedef std::vector<Charset>::const_iterator set_iterator;

       std::vector<Charset> pattern_;     // fixed sequence of characters each matching a
                                          // specified characterset

       FixedCharsetMatch(std::vector<Charset> const &m)
       : pattern_(m)
       {
       }


       virtual Charset finalCharSet() const
       {
           if(!pattern_.empty())
               return pattern_.back();

           return Charset();
       }

       virtual bool findFirstMatch(const_iterator &f, const_iterator &l) const
       {
           const_iterator scan = f;

           if(pattern_.empty())
              return false;

           Charset const &firstSet = pattern_[0];

           while(scan != l)
           {
               if(firstSet.test(*scan))
               {
                   // we have found a location where the first entry in the pattern is found

                   // now check for a complete match

                   const_iterator later = scan;

                   ++later;

                   set_iterator member = pattern_.begin(), final = pattern_.end();
                   ++member;

                   while(   (later != l)
                         && (member !=final)
                        )
                   {
                       ++later;
                       ++member;
                   }

                   if(member == final)
                   {
                       // got it!

                       f = scan;
                       l = later;

                       return true;

                   }

                   // if we get here, we did not find a match

               }
               else
                   ++scan;
           }


           return false;
       }


       virtual bool skipMatchingRegion(MatchType       t, 
                                       const_iterator &f, 
                                       const_iterator &l,
                                       bool            first,
                                       bool            last
                                      ) const
       {
           const_iterator scan;


           if(!searchBackward(t))
           {
               // standard forward search

               scan = f;
    
               set_iterator m = pattern_.begin();
               set_iterator e = pattern_.end();
    
               while(scan != l && m != e)
               {
                  char           cin  = *scan;
                  Charset const &cur  = *m;
    
                  if( !cur.test(cin) )
                     return false;
    
                  ++scan;
                  ++m;
               }
    
               if(
                    (m != e)
                 )
                  return false;
               f = scan;
           }
           else
           {
               // backwards search

               scan = l;
    
               set_iterator m = pattern_.begin();
               set_iterator e = pattern_.end();
    
               while(scan != f && e != m)
               {
                  --scan;
                  --e;

                  char           cin  = *scan;
                  Charset const &cur  = *e;
    
                  if( !cur.test(cin) )
                     return false;
    
               }
    
               if(
                    (m != e)
                 )
                  return false;

               l = scan;
           }

           return true;
       }

    };

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    PatternMatcher::~PatternMatcher()
    {
        for(size_t i = 0; i < matchExpression_.size();  ++i)
           delete matchExpression_[i];
    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////


    void PatternMatcher::addLiteralTerm(std::string &tmp)
    {
        if(!tmp.empty())
        {
           matchExpression_.push_back( new FixedStringMatch(tmp) );

           tmp.resize(0);

        }
    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////


    void PatternMatcher::parsePattern(std::string const &pattern)
    {
         if(pattern.empty())
         {
            return;
         }

         std::string::const_iterator f = pattern.begin(), l = pattern.end();

         std::string tmp;

         while( f != l )
         {
             char c = *f;

             if(c == '*')
             {
                 addLiteralTerm(tmp);  // if needed

                 while(    (f != l)
                        && (*f == '*') 
                      )
                 {
                    ++f;
                 }

                 if(f == l)
                 {
                    if(matchExpression_.empty())
                         matchExpression_.push_back( new ZeroOrMoreMatch() );
                    else
                         matchExpression_.push_back( new ZeroOrMoreMatch(true, 
                                                                         matchExpression_.back()
                                                                               ->finalCharSet()
                                                                        ) 
                                                   );


                 }
                 else
                 if(*f == '[')
                 {
                     // parse the character range expression that follows the *
                     // but leave f in place so that it can be used
                     // to populate next expression. 
                     //
                     // We will feed the parsed character set to the ZeroOrMore
                     // character scanner as the end criteria.

                     std::string::const_iterator scan = f, 
                                                 end   = l,
                                                 start
                                                 ;
                     
                     ++scan;

                     start = scan;

                     std::string tmp;

                     while(scan != l)
                     {
                        char c = *scan;

                        if(c == ']')
                        {
                           end = scan;
                           break;
                        }
                        else
                        if(c == '\\')
                        {
                           tmp += '\\';

                           ++scan;

                           if(scan == l)
                           {
                              break;
                           }
                           
                           tmp += *scan;
                        }
                        else
                          tmp += c;

                        ++scan;
                     }

                     // do not increment f here!

                     Charset set;

                     set.parse(start,end);

                     if(!matchExpression_.empty())
                         matchExpression_.push_back( new ZeroOrMoreMatch(set,
                                                                         matchExpression_.back()->finalCharSet()
                                                                        ) 
                                                   ); 
                     else
                         matchExpression_.push_back( new ZeroOrMoreMatch(set) ); 

                     // do not increment f here!

                 }
                 else
                 {
                    // single character set (or ?)

                    // do not increment f here!

                    if(!matchExpression_.empty())
                    {
                         Charset set;

                         if(*f == '?')
                             set.setAll();
                         else
                             set.set(*f, true);

                         matchExpression_.push_back( new ZeroOrMoreMatch(set,
                                                                         matchExpression_.back()->finalCharSet()
                                                                        ) 
                                                   ); 
                    }
                    else
                         matchExpression_.push_back( new ZeroOrMoreMatch(*f) ); 
                         
                    // do not increment f here!
                 }

                 continue; // f has been incremented

             }
             else
             if(c == '?')
             {
                 addLiteralTerm(tmp);  // if needed

                 size_t count=0;

                 while(f != l)
                 {
                    if(*f == '?')
                    {
                       ++count;
                    }
                    else
                        break;

                    ++f;
                 }


                 matchExpression_.push_back( new AnyExistingCharMatch(count) );

                 continue; // f has been increment

             }
             else
             if(c == '[')
             {
                 addLiteralTerm(tmp);  // if needed

                 ++f;  // skip [


                 std::vector<Charset> expr;

                 for(;;)
                 {
                     std::string tmp;

                     while(f != l && *f != ']')
                     {
                        c = *f;
                     
                        if(c == ']')
                        {
                           ++f;
                           break;
                        }
                        else
                        if(c == '\\')
                        {
                           ++f;
                     
                           if(f == l)
                              tmp += '\\';
                           else
                           {
                              tmp += c;   // leave the escape char in, unchanged
                              
                              tmp += *f;  // include the esscaped character
                              ++f;       

                           }
                        }
                        else
                          {
                            tmp += c;
                            ++f;
                          }
                     }

                     if(!tmp.empty())
                     {
                        expr.push_back(Charset());

                        expr.back().parse(tmp);
                     }

                     if( (f != l) && (*f == ']') )
                         ++f;

                     if(     (f == l)
                         ||  (*f != '[')
                       )
                        break;

                     // *f is [

                     ++f;

                 }

                 matchExpression_.push_back(new FixedCharsetMatch(expr) );

                 continue; // f has been incremented

             }
             else
             if(c == '\\')
             {
                 ++f;
                 if(f == l)
                   {
                    tmp += '\\';
                    continue; // ff is already incremented
                   }
                 else
                 if(*f == '0')
                 {
                     // process an octal number

                     char digit;
                     unsigned char value=0;
                     int count=0;

                     while(    (count < 3)
                            && (f != l) 
                          )
                     {
                         digit = *f;

                         if(digit >= '0' && digit <= '7')
                         {
                           value <<= 3;  count ++;
                            value += digit - '0';
                            ++f;
                            
                         }
                         else
                           { --f; break; } // push the character back into the stream

                     }

                     tmp += char(value);

                     
                 }
                 else
                    tmp += Charset::mapEscapeChar(*f);
             }
             else
             {
                tmp += c;
             }

             ++f;
         }

         addLiteralTerm(tmp);


    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////


    bool PatternMatcher::matches(std::string const &comparand) const
    {
       if(matchExpression_.empty())
           return false;

       const_iterator nc = comparand.begin(), lc = comparand.end();
       
       
       MatchExpression::const_iterator f = matchExpression_.begin(), l = matchExpression_.end();

       bool isFirst = true;

       while(f !=l)
       {
          MatchExpression::const_iterator tmp = f;
          ++tmp;

          bool matched = (*f)->skipMatchingRegion(EXACT, nc, lc, isFirst, tmp == l);

          isFirst=false;

          if(!matched)
              return false;

          ++f;
       }

       if(nc != lc)
          return false;    // the patterns must complete consume the string


       return true;
    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    size_t PatternMatcher::shortestLeadingMatch(std::string const &comparand) const
    {
        return leadingMatch(SHORTEST_LEADING, comparand);
    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    size_t PatternMatcher::longestLeadingMatch(std::string const &comparand) const
    {
        return leadingMatch(LONGEST_LEADING, comparand);
    }


    size_t PatternMatcher::leadingMatch(MatchType t, const_iterator &nc, const_iterator &lc) const
    {

       if(matchExpression_.empty())
           return false;

       MatchExpression::const_iterator f = matchExpression_.begin(), l = matchExpression_.end();

       const_iterator start = nc;

       bool isFirst = true;

       while(f !=l)
       {
          MatchExpression::const_iterator tmp = f;

          ++tmp;

          bool matched = (*f)->skipMatchingRegion(t, nc, lc, isFirst, tmp == l );

          isFirst = false;

          if(!matched)
              return 0;

          ++f;
       }

       // if we get here, we know that all the pattern matchers have skipped past the
       // leading parts of comparand that match the pattern and none of them failed
       // to match.

       return nc - start;  // length of the matched region.


    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    size_t PatternMatcher::shortestTrailingMatch(std::string const &comparand) const
    {
        return trailingMatch(SHORTEST_TRAILING, comparand);
    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    size_t PatternMatcher::longestTrailingMatch(std::string const &comparand) const
    {
        return trailingMatch(LONGEST_TRAILING, comparand);
    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    size_t PatternMatcher::trailingMatch(MatchType t, std::string const &comparand) const
    {

       // find the shortest match at the end of the string.
       //
       // This means scanning the string backwards and also scanning the patterns
       // backwards.


       if(matchExpression_.empty())
           return false;

       MatchExpression::const_iterator f = matchExpression_.begin(), l = matchExpression_.end();

       const_iterator nc = comparand.begin(), lc = comparand.end(), start;

       start = lc;

       bool isLast = true;

       while(l != f)
       {
          --l;

          bool matched = (*l)->skipMatchingRegion(t, nc, lc, l == f, isLast );

          isLast = false;

          if(!matched)
              return 0;

       }

       // if we get here, we know that all the pattern matchers have skipped past the
       // leading parts of comparand that match the pattern and none of them failed
       // to match.

       return start - lc;  // length of the matched region.   (starting at 'start' and working backwards)


    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    std::string PatternMatcher::removeLeadingMatch(std::string const &s, bool longestMatch) const
    {
        std::string rv(s);

        size_t      length = longestMatch ? longestLeadingMatch(rv) : shortestLeadingMatch(rv);
        size_t      total  = rv.size();
        
        if(length == total)
        {
            rv.resize(0);
        }
        else
        if(length < total)
        {
           rv = rv.substr(length, total - length);
        }

        return rv;
        
    }

    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////

    std::string PatternMatcher::removeTrailingMatch(std::string const &s, bool longestMatch) const
    {
        std::string rv(s);

        size_t length = longestMatch ? longestTrailingMatch(rv) : shortestTrailingMatch(rv);
        size_t total  = rv.size();
        
        if(length < total)
        {
           rv = rv.substr(0, total - length);
        }
        else
        if(length == total)
        {
           rv.resize(0);
        }

        return rv;

    }


    PatternMatcher::OffsetAndLength PatternMatcher::find(std::string const &comparand, size_t startPos) const
    {
       if(matchExpression_.empty())
           return OffsetAndLength(comparand.empty() ? 0 : std::string::npos,0);  // nothing matches nothing, else can't find nothing in something

       if(comparand.empty())
           return OffsetAndLength(std::string::npos,0);  // you can't find something in nothing.

       if(startPos >= comparand.size())
           return OffsetAndLength(std::string::npos,0);

       const_iterator nc = comparand.begin(), lc = comparand.end(), start;

       nc += startPos;

       Matcher const &first = *matchExpression_.front();

       while(nc != lc)
       {
           const_iterator tmp = lc;

           if(!first.findFirstMatch(nc, tmp))
           {
                break;  // no more matches possible
           }

           // we found at least the first sub-expression, somewhere.  tmp
           // has been moved back from l to the end of the first match at this
           // point.  nc points to the start of the first match

           const_iterator tmp2 = nc;
 
           tmp = lc;  // tmp was moved by the findFirstMatch call above, we are
                      // using tmp vars because these functions touch both first
                      // and last iterators

           size_t matchLength;

           if(0 != (matchLength = leadingMatch(LONGEST_LEADING, tmp2, tmp)) )
           {
              // got it, we're done

              return OffsetAndLength(nc - comparand.begin(),
                                     matchLength
                                    );

           }

           // nc is partial match but no cookie.

           ++nc;
       }



       return OffsetAndLength(std::string::npos,0);

    }

   std::string 
   PatternMatcher::replaceMatches(std::string const &s, std::string const &r, bool allMatches) const
   {
        std::string rv;

        size_t nextSearchPoint = 0;

        size_t length = s.size();

        for(;;)
        {
           // replace one or all of the matches of *this in s with r.

           OffsetAndLength loc = find(s, nextSearchPoint);

           if(loc.first >= length)
           {
              // there are no more matches in s

              rv.append(  s.begin() + nextSearchPoint, s.end() );

              break;

           }

           // we found a match somewhere after nextSearchPoint

           if(loc.first > nextSearchPoint)
           {
              rv.append(  s.begin() + nextSearchPoint,
                          s.begin() + nextSearchPoint + (loc.first - nextSearchPoint)
                       );
           }


           rv += r;


           nextSearchPoint = loc.first + loc.second;

           if(nextSearchPoint >= length)
               break;

           if(!allMatches)
           {
               if(nextSearchPoint != length)
               {
                  rv.append(s.begin()+nextSearchPoint,
                            s.end()
                           );
               }
               break;
           }

        }



        return rv;
   }



};
